# MIT License

# PointCNN
# Copyright (c) 2018 Shandong University
# Copyright (c) 2018 Yangyan Li, Rui Bu, Mingchao Sun, Baoquan Chen

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import glob
import importlib
import sys
import warnings
import os
import math
from pathlib import Path
import json
import types
import random
import logging
import shutil
logger = logging.getLogger()

try:
    from torch.utils.data import DataLoader, Dataset, SubsetRandomSampler
    import torch.nn.functional as F
    import torch
    import numpy as np
    from fastai.data_block import DataBunch
    import matplotlib.pyplot as plt
    from mpl_toolkits.mplot3d import Axes3D
    import arcgis
    from fastai.data_block import ItemList
    from fastprogress.fastprogress import master_bar, progress_bar
    from scipy.spatial.transform import Rotation as R
except ImportError:
    # To avoid breaking builds.
    class Dataset():
        pass
    class ItemList():
        pass

def try_imports(list_of_modules):
    ## Not a generic function.
    try:
        for module in list_of_modules:
            importlib.import_module(module)
    except Exception as e:
        raise Exception(f"""This function requires {' '.join(list_of_modules)}. Visit https://developers.arcgis.com/python/guide/install-and-set-up and https://developers.arcgis.com/python/guide/point-cloud-segmentation-using-pointcnn for installing the dependencies.""")

def try_import(module):
    try:
        importlib.import_module(module)
    except ModuleNotFoundError:
        if module == 'plotly':
            raise Exception("This function requires plotly. Visit https://developers.arcgis.com/python/guide/install-and-set-up and https://developers.arcgis.com/python/guide/point-cloud-segmentation-using-pointcnn for installing the dependencies.")
        elif module == 'laspy':
            raise Exception("This function requires laspy. Visit https://developers.arcgis.com/python/guide/install-and-set-up and https://developers.arcgis.com/python/guide/point-cloud-segmentation-using-pointcnn for installing the dependencies.")
        elif module == 'h5py':
            raise Exception(f"This function requires h5py. Visit https://developers.arcgis.com/python/guide/install-and-set-up and https://developers.arcgis.com/python/guide/point-cloud-segmentation-using-pointcnn for installing the dependencies.")
        else:
            raise Exception(f"This function requires {module}. Please install it in your environment.")

def pad_tensor(cur_tensor, max_points, to_float=True):
    cur_points = cur_tensor.shape[0]
    if cur_points < max_points:
        remaining_points = max_points - cur_points
        if len(cur_tensor.shape) < 2:
            remaining_tensor = torch.zeros(remaining_points)
        else:
            remaining_tensor = torch.zeros(remaining_points, cur_tensor.shape[1])
        if to_float:
            remaining_tensor = remaining_tensor.float()
        else:
            remaining_tensor = remaining_tensor.long()
        cur_tensor = torch.cat((cur_tensor, remaining_tensor), dim=0)
    else:
        cur_tensor = cur_tensor[:max_points]
        cur_points = max_points 
    return cur_tensor, cur_points

def concatenate_tensors(read_file, input_keys, tile, max_points):
    cat_tensor = []
    
    cur_tensor = torch.tensor(read_file['xyz'][tile[1]:tile[1]+tile[2]].astype(np.float32))
    if len(cur_tensor.shape) < 2:
        cur_tensor = cur_tensor[:, None]

    cur_tensor, cur_points = pad_tensor(cur_tensor, max_points)
    cat_tensor.append(cur_tensor)
    
    for key, min_max in input_keys.items():
        if key not in ['xyz']:
            cur_tensor = torch.tensor(read_file[key][tile[1]:tile[1]+tile[2]].astype(np.float32))
            if len(cur_tensor.shape) < 2:
                cur_tensor = cur_tensor[:, None]

            max_val = cur_tensor.new(min_max['max'])
            min_val = cur_tensor.new(min_max['min'])
            cur_tensor = cur_tensor = (cur_tensor - min_val) / (max_val )  ## Test with one_hot
            cur_tensor, cur_points = pad_tensor(cur_tensor, max_points)
            cat_tensor.append(cur_tensor)
    
    return torch.cat(cat_tensor, dim=1), cur_points

def remap_classes(class_values):
    flag = False
    if class_values[0] != 0:
        return True
    for i in range(len(class_values) - 1):
        if class_values[i] + 1 != class_values[i+1]:
            return True
    return flag    

class PointCloudDataset(Dataset):
    def __init__(self, path, max_point, extra_dim, class_mapping, **kwargs):
        try_import("h5py")
        import h5py
        self.init_kwargs = kwargs
        self.path = Path(path)
        self.max_point = max_point  ## maximum number of points
        
        with open(self.path / 'Statistics.json', 'r') as f:
            self.statistics = json.load(f)        
        
        self.block_size = self.statistics['parameters']['tileSize']
        self.input_keys = self.statistics['features']  ## Keys to include in training
        self.classification_key = kwargs.get('classification_key', 'classification')     ## Key which contain labels
        self.extra_dim = extra_dim
        self.total_dim = 3 + extra_dim
        self.extra_features = self.input_keys
        
        if class_mapping is None:
            self.class_mapping =  {value['classCode']:idx for idx,value in enumerate(self.statistics['classification']['table'])}
        else:
            self.class_mapping = class_mapping
        ## Helper attributes for remapping
        self.c = len(self.class_mapping)
        self.color_mapping = kwargs.get('color_mapping', np.array([np.random.randint(0, 255, 3) for i in range(self.c)])/255)
        self.remap = False
        self.remap = remap_classes(list(self.class_mapping.values()))
        
        self.classes = list(self.class_mapping.values())

        with h5py.File(self.path / 'ListTable.h5', 'r') as f:
            files = f['Files'][:]
            self.tiles = f['Tiles'][:]
        
        self.relative_files = files
        self.filenames = [self.path / file.decode() for file in files]
        self.h5files = [(h5py.File(filename, 'r')) for filename in self.filenames]
    
    def __len__(self):
        return len(self.tiles)
    
    def __getitem__(self, i):
        
        tile = self.tiles[i]        
        read_file = self.h5files[tile[0]]
        
        if self.classification_key in read_file.keys():
            
            classification, _ = pad_tensor(torch.tensor(read_file[self.classification_key][tile[1]:tile[1]+tile[2]].astype(int)),
                                        self.max_point,
                                        to_float=False
                                        )
            if not self.remap:
                return concatenate_tensors(read_file, self.input_keys, tile, self.max_point), classification
            else:
                return concatenate_tensors(read_file, self.input_keys, tile, self.max_point), remap_labels(classification)                                        
        else:
            logger.warning(f"key `{self.classification_key}` could not be found in the exported files.")
            return concatenate_tensors(read_file, self.input_keys, tile, self.max_point), None
        
    def close(self):
        [file.close() for file in self.h5files]

def minmax_scale(pc):
    min_val = np.amin(pc, axis=0)
    max_val = np.amax(pc, axis=0)
    return (pc - min_val[None])/max(max_val - min_val)

def recompute_color_mapping(color_mapping, all_classes):
    color_mapping = {int(k):v for k, v in color_mapping.items()}
    try:
        color_mapping = {k:color_mapping[k] for k in all_classes}
    except KeyError:
        raise Exception(f"Keys of your classes in your color_mapping do not match with classes present in data i.e {all_classes}")
    return color_mapping

def class_string(label_array, prefix=''):
    return [f'{prefix}class: {k}' for k in label_array]

def mask_classes(labels, mask_class, class_mapping=None):
    if class_mapping is not None:
        mask_class = [class_mapping[x] for x in mask_class]
    if mask_class == []:
        ## return complete mask
        return labels != None
    else:
        sample_idxs = np.concatenate([(labels[None]!=mask) for mask in mask_class])
        sample_idxs = sample_idxs.all(axis=0)
        return sample_idxs    

def get_max_display_points(self, kwargs):
    if "max_display_point" in kwargs.keys():
        max_display_point = kwargs['max_display_point']
        self.max_display_point = max_display_point
    else:
        if hasattr(self, 'max_display_point'):
            max_display_point = self.max_display_point
        else:
            max_display_point = 20000   
    return max_display_point

def show_point_cloud_batch(self, rows=2, figsize=(6,12), color_mapping=None, **kwargs):

    """
    It will plot 3d point cloud data you exported in the notebook.

    =====================   ===========================================
    **Argument**            **Description**
    ---------------------   -------------------------------------------
    rows                    Optional rows. Number of rows to show. Default
                            value is 2 and maximum value is the `batch_size`
                            passed in `prepare_data`. 
    ---------------------   -------------------------------------------
    color_mapping           Optional dictionary. Mapping from class value
                            to RGB values. Default value example:
                            {0:[220,220,220], 2:[255,0,0], 6:[0,255,0]}.
    =====================   ===========================================

    **kwargs**

    =====================   ===========================================
    **Argument**            **Description**
    ---------------------   -------------------------------------------
    mask_class              Optional list of integers. Array containing
                            class values to mask. Use this parameter to 
                            display the classes of interest.
                            Default value is []. 
                            Example: All the classes are in [0, 1, 2]
                            to display only class `0` set the mask class
                            parameter to be [1, 2]. List of all classes
                            can be accessed from `data.classes` attribute
                            where `data` is the `Databunch` object returned
                            by `prepare_data` function.    
    ---------------------   -------------------------------------------
    width                   Optional integer. Width of the plot. Default 
                            value is 750.
    ---------------------   -------------------------------------------
    height                  Optional integer. Height of the plot. Default
                            value is 512.
    ---------------------   -------------------------------------------
    max_display_point       Optional integer. Maximum number of points
                            to display. Default is 20000. A warning will
                            be raised if the total points to display exceeds
                            this parameter. Setting this parameter will
                            randomly sample the specified number of points
                            and once set, it will be used for future uses.                               
    =====================   ===========================================
    """

    filter_outliers = False
    try_import("h5py")
    import h5py
    try_import('plotly')
    import plotly.graph_objects as go
    mask_class = kwargs.get('mask_class', [])
    max_display_point = get_max_display_points(self, kwargs)
    rows = min(rows, self.batch_size)
    color_mapping = self.color_mapping if color_mapping is None else color_mapping
    color_mapping = recompute_color_mapping(color_mapping, self.classes)       
    color_mapping = np.array(list(color_mapping.values())) / 255

    h5_files = self.h5files.copy()
    random.shuffle(h5_files)  

    idx = 0
    file_idx = 0
    while (idx < rows):
        file = h5_files[file_idx]
        pc = file['xyz'][:]
        labels = file['classification'][:]
        unmapped_labels = labels.copy()
        if self.remap:
            labels = remap_labels(labels, self.class_mapping) 
        sample_idxs = mask_classes(labels, mask_class, self.class_mapping if self.remap else None)
        sampled_pc = pc[sample_idxs]

        if sampled_pc.shape[0] == 0:
            file_idx += 1
            continue
        x, y, z = recenter(sampled_pc).transpose(1,0)  ## convert to 3,N so that upacking works

        if filter_outliers:
            ## Filter on the basis of std.
            mask = filter_pc(pc)
        else:
            ## all points
            mask = x > -9999999    

        if sample_idxs.sum() > max_display_point:
            raise_maxpoint_warning(idx, kwargs, logger, max_display_point)
            mask = np.random.randint(0, sample_idxs.sum(), max_display_point)   
        
        color_list = color_mapping[labels[sample_idxs]][mask].tolist()
        scene=dict(aspectmode='data')

        layout = go.Layout(
            width=kwargs.get('width', 750),
            height=kwargs.get('height', 512),
            scene = scene)

        fig = go.Figure(data=[go.Scatter3d(x=x[mask], y=y[mask], z=z[mask], 
                                        mode='markers', marker=dict(size=1, color=color_list),
                                        text=class_string(unmapped_labels[sample_idxs][mask]))], layout=layout)
        fig.show()        

        if idx == rows-1:
            break        
        idx += 1
        file_idx += 1

def filter_pc(pc):
    mean = pc.mean(0)
    std = pc.std(0)
    mask = (pc[:, 0] < (mean[0] + 2*std[0])) & (pc[:, 1] < (mean[1] + 2*std[1])) & (pc[:, 2] < (mean[2] + 2*std[2]))
    return mask

def recenter(pc):
    min_val = np.amin(pc, axis=0)
    max_val = np.amax(pc, axis=0)
    return (pc - min_val[None])

def show_point_cloud_batch_TF(self, rows=2, color_mapping=None, **kwargs):

    """
    It will plot 3d point cloud data you exported in the notebook.

    =====================   ===========================================
    **Argument**            **Description**
    ---------------------   -------------------------------------------
    rows                    Optional rows. Number of rows to show. Default
                            value is 2 and maximum value is the `batch_size`
                            passed in `prepare_data`. 
    ---------------------   -------------------------------------------
    color_mapping           Optional dictionary. Mapping from class value
                            to RGB values. Default value example:
                            {0:[220,220,220], 2:[255,0,0], 6:[0,255,0]}.
    =====================   ===========================================

    **kwargs**

    =====================   ===========================================
    **Argument**            **Description**
    ---------------------   -------------------------------------------
    mask_class              Optional list of integers. Array containing
                            class values to mask. Use this parameter to 
                            display the classes of interest.
                            Default value is []. 
                            Example: All the classes are in [0, 1, 2]
                            to display only class `0` set the mask class
                            parameter to be [1, 2]. List of all classes
                            can be accessed from `data.classes` attribute
                            where `data` is the `Databunch` object returned
                            by `prepare_data` function.    
    ---------------------   -------------------------------------------
    width                   Optional integer. Width of the plot. Default 
                            value is 750.
    ---------------------   -------------------------------------------
    height                  Optional integer. Height of the plot. Default
                            value is 512.
    ---------------------   -------------------------------------------
    max_display_point       Optional integer. Maximum number of points
                            to display. Default is 20000. A warning will
                            be raised if the total points to display exceeds
                            this parameter. Setting this parameter will
                            randomly sample the specified number of points
                            and once set, it will be used for future uses.                                                                                
    =====================   ===========================================
    """

    filter_outliers = False
    try_import("h5py")
    import h5py
    try_import('plotly')
    import plotly.graph_objects as go
    mask_class = kwargs.get('mask_class', [])
    rows = min(rows, self.batch_size)
    max_display_point = get_max_display_points(self, kwargs)
    color_mapping = self.color_mapping if color_mapping is None else color_mapping
    color_mapping = recompute_color_mapping(color_mapping, self.classes)       
    color_mapping = np.array(list(color_mapping.values())) / 255

    idx = 0
    import random
    keys = list(self.meta['files'].keys()).copy()
    keys = [k for k in keys if 'train' in Path(k).parts]
    random.shuffle(keys)
    
    for idx_file, fn in enumerate(keys):
        num_files = self.meta['files'][fn]['idxs']
        block_center = self.meta['files'][fn]['block_center']
        block_center = np.array(block_center)
        block_center[0][2], block_center[0][1] = block_center[0][1], block_center[0][2]
        if num_files == []:
            continue
        if not Path(fn).is_absolute():
            fn = str(self.path / fn)
        idxs = [h5py.File(fn[:-3] + f'_{i}.h5', 'r') for i in num_files]
        pc = []
        labels = []
        for i in idxs:
            current_block = i['unnormalized_data'][:, :3]
            data_num = i['data_num'][()]   
            pc.append(current_block[:data_num])
            labels.append(i['label_seg'][:data_num])
            i.close()
            
        if pc == []:
            continue         
       
        pc = np.concatenate(pc, axis=0)
        labels = np.concatenate(labels, axis=0)
        unmapped_labels = labels.copy()    
        if self.remap:
            labels = remap_labels(labels, self.class_mapping)  
        sample_idxs = mask_classes(labels, mask_class, self.class_mapping if self.remap else None)
        sampled_pc = pc[sample_idxs]
        if sampled_pc.shape[0] == 0:
            continue
        x, y, z = recenter(sampled_pc).transpose(1,0)       
        if filter_outliers:
            ## Filter on the basis of std.
            mask = filter_pc(pc)
        else:
            ## all points
            mask = [True] * len(x)

        if sample_idxs.sum() > max_display_point:
            raise_maxpoint_warning(idx_file, kwargs, logger, max_display_point)
            mask = np.random.randint(0, sample_idxs.sum(), max_display_point)  
            
        color_list =  color_mapping[labels[sample_idxs]][mask].tolist()
        
        scene=dict(aspectmode='data')
        layout = go.Layout(
            width=kwargs.get('width', 750),
            height=kwargs.get('height', 512),
            scene = scene)

        figww = go.Figure(data=[go.Scatter3d(x=x[mask], y=z[mask], z=y[mask], 
                                        mode='markers', marker=dict(size=1, color=color_list),
                                        text=class_string(unmapped_labels[sample_idxs][mask]))], layout=layout)
        figww.show()

        if idx == rows-1:
            break
        idx += 1

def get_device():
    if getattr(arcgis.env, "_processorType", "") == "GPU" and torch.cuda.is_available():
        device = torch.device("cuda")
    elif getattr(arcgis.env, "_processorType", "") == "CPU":
        device = torch.device("cpu")
    else:
        device = torch.device('cuda') if torch.cuda.is_available() else torch.device('cpu')

    return device    

def read_xyzinumr_label_from_las(filename_las, extra_features):
    try_import('laspy')
    import laspy
    file = laspy.file.File(filename_las, mode='r')    
    h = file.header
    xyzirgb_num = h.point_records_count
    labels = file.Classification
    
    xyz = np.concatenate([file.x[:, None], file.y[:, None], file.z[:, None]] + [(np.clip(getattr(file, f[0]), None, f[1])[:, None] - f[2])/ (f[1] - f[2]) for f in extra_features],
                         axis=1)
    
    xyzirgb_num = len(xyz)
    file.close()
    return xyz, labels, xyzirgb_num

def prepare_las_data(root,
                     block_size,
                     max_point_num,
                     output_path,
                     extra_features=[('intensity', 5000, 0), ('num_returns', 5, 0)],
                     grid_size=1.0,
                     blocks_per_file=2048,
                     folder_names=['train', 'val'],
                     segregate=True,
                     **kwargs
                    ):
    try_import("h5py")
    import h5py
    block_size_ = block_size
    batch_size= blocks_per_file
    data = np.zeros((batch_size, max_point_num, 3 + len(extra_features))) #XYZ, Intensity, NumReturns
    unnormalized_data = np.zeros((batch_size, max_point_num, 3 + len(extra_features)))
    data_num = np.zeros((batch_size), dtype=np.int32)
    label = np.zeros((batch_size), dtype=np.int32)
    label_seg = np.zeros((batch_size, max_point_num), dtype=np.int32)
    indices_split_to_full = np.zeros((batch_size, max_point_num), dtype=np.int32)
    LOAD_FROM_EXT = '.las'
    os.makedirs(output_path, exist_ok=True)

    if (Path(output_path) / 'meta.json').exists() or (Path(output_path) / 'Statistics.json').exists():
        raise Exception(f"The given output path({output_path}) already contains exported data. Either delete those files or pass in a new output path.")

    folders = [os.path.join(root, folder) for folder in folder_names]  ## Folders are named train and val
    mb = master_bar(range(len(folders)))
    for itn in mb:
        folder = folders[itn]
        os.makedirs(os.path.join(output_path, Path(folder).stem), exist_ok=True)
        datasets = [filename[:-4] for filename in os.listdir(folder) if filename.endswith(LOAD_FROM_EXT)]
        # mb.write(f'{itn + 1}. Exporting {Path(folder).stem} folder')
        for dataset_idx, dataset in enumerate(progress_bar(datasets, parent=mb)):
            filename_ext = os.path.join(folder, dataset + LOAD_FROM_EXT)
            if LOAD_FROM_EXT == '.las':
                xyzinumr, labels, xyz_num = read_xyzinumr_label_from_las(filename_ext, extra_features)
                xyz, other_features = np.split(xyzinumr, (3,), axis=-1)
                if len(other_features.shape) < 2:
                    other_features = other_features[:, None]
            else:
                xyz, labels, xyz_num = read_xyz_label_from_txt(filename_ext)
            
            offsets = [('zero', 0.0), ('half', block_size_ / 2)]

            for offset_name, offset in offsets:
                idx_h5 = 0
                idx = 0

                xyz_min = np.amin(xyz, axis=0, keepdims=True) - offset
                xyz_max = np.amax(xyz, axis=0, keepdims=True)
                block_size = (block_size_, block_size_, 2 * (xyz_max[0, -1] - xyz_min[0, -1]))  
                xyz_blocks = np.floor((xyz - xyz_min) / block_size).astype(np.int)

                blocks, point_block_indices, block_point_counts = np.unique(xyz_blocks, return_inverse=True,
                                                                            return_counts=True, axis=0)
                block_point_indices = np.split(np.argsort(point_block_indices), np.cumsum(block_point_counts[:-1]))

                block_to_block_idx_map = dict()
                for block_idx in range(blocks.shape[0]):
                    block = (blocks[block_idx][0], blocks[block_idx][1])
                    block_to_block_idx_map[(block[0], block[1])] = block_idx

                # merge small blocks into one of their big neighbors
                block_point_count_threshold = max_point_num / 10
                nbr_block_offsets = [(0, 1), (1, 0), (0, -1), (-1, 0), (-1, 1), (1, 1), (1, -1), (-1, -1)]
                block_merge_count = 0
                for block_idx in range(blocks.shape[0]):
                    if block_point_counts[block_idx] >= block_point_count_threshold:
                        continue

                    block = (blocks[block_idx][0], blocks[block_idx][1])
                    for x, y in nbr_block_offsets:
                        nbr_block = (block[0] + x, block[1] + y)
                        if nbr_block not in block_to_block_idx_map:
                            continue

                        nbr_block_idx = block_to_block_idx_map[nbr_block]
                        if block_point_counts[nbr_block_idx] < block_point_count_threshold:
                            continue

                        block_point_indices[nbr_block_idx] = np.concatenate(
                            [block_point_indices[nbr_block_idx], block_point_indices[block_idx]], axis=-1)
                        block_point_indices[block_idx] = np.array([], dtype=np.int)
                        block_merge_count = block_merge_count + 1
                        break

                idx_last_non_empty_block = 0
                for block_idx in reversed(range(blocks.shape[0])):
                    if block_point_indices[block_idx].shape[0] != 0:
                        idx_last_non_empty_block = block_idx
                        break

                # uniformly sample each block
                for block_idx in range(idx_last_non_empty_block + 1):
                    point_indices = block_point_indices[block_idx]
                    if point_indices.shape[0] == 0:
                        continue
                    block_points = xyz[point_indices]
                    block_min = np.amin(block_points, axis=0, keepdims=True)
                    xyz_grids = np.floor((block_points - block_min) / grid_size).astype(np.int)
                    grids, point_grid_indices, grid_point_counts = np.unique(xyz_grids, return_inverse=True,
                                                                            return_counts=True, axis=0)
                    grid_point_indices = np.split(np.argsort(point_grid_indices), np.cumsum(grid_point_counts[:-1]))
                    grid_point_count_avg = int(np.average(grid_point_counts))
                    point_indices_repeated = []
                    for grid_idx in range(grids.shape[0]):
                        point_indices_in_block = grid_point_indices[grid_idx]
                        repeat_num = math.ceil(grid_point_count_avg / point_indices_in_block.shape[0])
                        if repeat_num > 1:
                            point_indices_in_block = np.repeat(point_indices_in_block, repeat_num)
                            np.random.shuffle(point_indices_in_block)
                            point_indices_in_block = point_indices_in_block[:grid_point_count_avg]
                        point_indices_repeated.extend(list(point_indices[point_indices_in_block]))
                    block_point_indices[block_idx] = np.array(point_indices_repeated)
                    block_point_counts[block_idx] = len(point_indices_repeated)
                for block_idx in range(idx_last_non_empty_block + 1):
                    point_indices = block_point_indices[block_idx]
                    if point_indices.shape[0] == 0:
                        continue

                    block_point_num = point_indices.shape[0]
                    block_split_num = int(math.ceil(block_point_num * 1.0 / max_point_num))
                    point_num_avg = int(math.ceil(block_point_num * 1.0 / block_split_num))
                    point_nums = [point_num_avg] * block_split_num
                    point_nums[-1] = block_point_num - (point_num_avg * (block_split_num - 1))
                    starts = [0] + list(np.cumsum(point_nums))

                    np.random.shuffle(point_indices)
                    block_points = xyz[point_indices]
                    block_min = np.amin(block_points, axis=0, keepdims=True)
                    block_max = np.amax(block_points, axis=0, keepdims=True)
                    block_center = (block_min + block_max) / 2
                    block_center[0][-1] = block_min[0][-1]
                    unnormalized_block_points = block_points.copy()
                    block_points = block_points - block_center  # align to block bottom center
                    x, y, z = np.split(block_points, (1, 2), axis=-1)
        
                    block_xzyrgbi = np.concatenate([x, z, y] + [i[point_indices][:, None] for i in other_features.transpose(1,0)], axis=-1) #XYZ, Intensity, NumReturns, RGB
                    block_labels = labels[point_indices]

                    ## unormalized points
                    x_u, y_u, z_u = np.split(unnormalized_block_points, (1, 2), axis=-1)
                    unnormalized_block_xzyrgbi = np.concatenate([x_u, z_u, y_u] + [i[point_indices][:, None] for i in other_features.transpose(1,0)], axis=-1)


                    for block_split_idx in range(block_split_num):
                        start = starts[block_split_idx]
                        point_num = point_nums[block_split_idx]
                        end = start + point_num
                        idx_in_batch = idx % batch_size
                        data[idx_in_batch, 0:point_num, ...] = block_xzyrgbi[start:end, :]
                        unnormalized_data[idx_in_batch, 0:point_num, ...] = unnormalized_block_xzyrgbi[start:end, :]
                        data_num[idx_in_batch] = point_num
                        label[idx_in_batch] = dataset_idx  # won't be used...
                        label_seg[idx_in_batch, 0:point_num] = block_labels[start:end]
                        indices_split_to_full[idx_in_batch, 0:point_num] = point_indices[start:end]

                        if ((idx + 1) % batch_size == 0) or \
                                (block_idx == idx_last_non_empty_block and block_split_idx == block_split_num - 1):
                            item_num = idx_in_batch + 1
                            filename_h5 = os.path.join(output_path, Path(folder).stem, dataset + '_%s_%d.h5' % (offset_name, idx_h5))
                            file = h5py.File(filename_h5, 'w')
                            file.create_dataset('unnormalized_data', data=unnormalized_data[0:item_num, ...])
                            file.create_dataset('data', data=data[0:item_num, ...])
                            file.create_dataset('data_num', data=data_num[0:item_num, ...])
                            file.create_dataset('label', data=label[0:item_num, ...])
                            file.create_dataset('label_seg', data=label_seg[0:item_num, ...])
                            file.create_dataset('indices_split_to_full', data=indices_split_to_full[0:item_num, ...])
                            file.create_dataset('block_center', data=block_center)
                            file.close()
                            idx_h5 = idx_h5 + 1
                        idx = idx + 1
    if segregate:
        ## Segregate data
        output_path = Path(output_path)
        path_convert = output_path
        
        GROUND_CLASS = 0
        mb = master_bar(range(len(folders)))
        meta_file = {}
        meta_file['files'] = {}
        all_classes = set()
        for itn in mb:
            folder = folders[itn]
            path = output_path / Path(folder).stem
            total = 0
            # mb.write(f'{itn + 1}. Segregating {Path(folder).stem} folder')
            all_files = list(path.glob('*.h5'))
            file_id = 0
            for idx, fn in enumerate(progress_bar(all_files, parent=mb)):
                file = h5py.File(fn, 'r')
                data = file['data']
                total += data.shape[0]
                label_seg = file['label_seg']
                data_num = file['data_num']
                unnormalized_data = file['unnormalized_data']
                block_center = file['block_center'][:]
                file_idxs = []
                for i in range(file['data_num'][:].shape[0]):
                    if not ((label_seg[i] != GROUND_CLASS).sum().tolist() == 0):
                        save_file = path_convert / Path(folder).stem /  (fn.stem + f'_{i}' + '.h5')
                        new_file = h5py.File(save_file, mode='w')
                        new_file.create_dataset('unnormalized_data', data=unnormalized_data[i])
                        new_file.create_dataset('data', data=data[i])
                        new_file.create_dataset('label_seg', data=label_seg[i])
                        new_file.create_dataset('data_num', data=data_num[i])
                        new_file.close()
                        all_classes = all_classes.union(np.unique(label_seg[i][:data_num[i]]).tolist())
                        file_idxs.append(i)
                meta_file['files'][os.path.join(*fn.parts[-2:])] = {'idxs':file_idxs,
                                                'block_center':block_center.tolist()}
                file.close()
                os.remove(fn)
        meta_file['num_classes'] = len(all_classes)
        meta_file['classes'] = list(all_classes)
        meta_file['max_point'] = max_point_num
        meta_file['num_extra_dim'] = len(extra_features)
        meta_file['extra_features'] = extra_features
        meta_file['block_size'] = block_size
        with open(output_path / 'meta.json', 'w') as f:
            json.dump(meta_file, f)

    if kwargs.get('print_it', True):
        print('Export finished.')

    return output_path

## Segregated data ItemList

def open_h5py_tensor(fn, keys=['data']):
    try_import("h5py")
    import h5py
    data_label = []
    file = h5py.File(fn, 'r')
    for key in keys:
        tensor = torch.tensor(file[key][...]).float()
        data_label.append(tensor)

    file.close()    
    return data_label ## While getting a specific index from the file
    
## It also stores the label so that we don't have to open the file twice.
class DataStore():
    pass
    
class PointCloudItemList(ItemList):
    def __init__(self, items, **kwargs):
        super().__init__(items, **kwargs)
        
        self.keys = ['data', 'label_seg', 'data_num']
        
    def get(self, i):
        data = self.open(self.items[i])
        DataStore.i = i
        DataStore.data = data
        return (data[0], data[2])
        
    def open(self, fn):
        return open_h5py_tensor(fn, keys=self.keys)

def remap_labels(labels, class_mapping):
    if isinstance(labels, torch.Tensor):
        remapped_label = torch.zeros_like(labels)
    else:
        remapped_label = np.zeros_like(labels)
    for k,v in class_mapping.items():
        remapped_label[labels == k] = v
    return remapped_label
    
class PointCloudLabelList(ItemList):
    def __init__(self, items, remap=False, class_mapping={}, **kwargs):
        super().__init__(items, **kwargs)
        self.key = 'label_seg'
        self.remap = remap
        self.class_mapping = class_mapping
        
    def get(self, i):
        if self.remap:
            # import pdb; pdb.set_trace();
            return remap_labels(DataStore.data[1].long(), self.class_mapping)
        else:
            return DataStore.data[1].long()
    
    def analyze_pred(self, pred):
        return pred.argmax(dim=1)
        
PointCloudItemList._label_cls = PointCloudLabelList

## Prepare data called in _data.py

def pointcloud_prepare_data(path, class_mapping, batch_size, val_split_pct, dataset_type='PointCloud', transform_fn=None, **kwargs):
    try_imports(['h5py', 'plotly', 'laspy'])
    databunch_kwargs = {'num_workers':0} if sys.platform == 'win32' else {}
    if (path / 'Statistics.json').exists():
        dataset_type = "PointCloud"
    elif (path / 'meta.json').exists():
        dataset_type = "PointCloud_TF"
    else:
        dataset_type = "Unknown"

    if dataset_type == 'PointCloud':
        with open(path / 'Statistics.json') as f:
            json_file = json.load(f)

        max_points = json_file['parameters']['numberOfPointsInEachTile']

        ## It is assumed that the pointcloud will have only X,Y & Z.
        extra_dim = sum([len(v['max']) if isinstance(v['max'], list) else 1 for k,v in json_file['features'].items()]) - 3
        pointcloud_dataset = PointCloudDataset(path, max_points, extra_dim, class_mapping, **kwargs)

        # Splitting in train and test based on files.
        total_files = len(pointcloud_dataset.filenames)
        total_files_idxs = list(range(total_files))
        random.shuffle(total_files_idxs)
        total_val_files = int(val_split_pct * total_files)
        val_files = total_files_idxs[-total_val_files:]
        if total_val_files == 0:
            raise Exception("No files could be added to validation dataset. Please increase the value of `val_split_pct`")
        tile_file_indices = pointcloud_dataset.tiles[:, 0]
        val_indices = torch.from_numpy(np.isin(tile_file_indices, val_files)).nonzero()
        train_indices = torch.from_numpy(np.logical_not(np.isin(tile_file_indices, val_files))).nonzero()
        train_sampler = SubsetRandomSampler(train_indices)
        val_sampler = SubsetRandomSampler(val_indices)
        train_dl = DataLoader(pointcloud_dataset, batch_size=batch_size, sampler=train_sampler, **databunch_kwargs)
        valid_dl = DataLoader(pointcloud_dataset, batch_size=batch_size, sampler=val_sampler, **databunch_kwargs)
        device = get_device()
        data = DataBunch(train_dl, valid_dl, device=device)
        data.show_batch = types.MethodType(show_point_cloud_batch, data)
        data.path = data.train_ds.path
        data.val_files = val_files

    elif dataset_type == 'PointCloud_TF':
        with open(Path(path) / 'meta.json', 'r') as f:
            meta = json.load(f)
        classes = meta['classes']
        remap = remap_classes(classes)
        class_mapping = class_mapping if class_mapping is not None else {v:k for k,v in enumerate(classes)}
        src = PointCloudItemList.from_folder(path, ['.h5'])
        train_idxs = [i for i,p in enumerate(src.items) if p.parent.name == 'train']
        val_idxs = [i for i,p in enumerate(src.items) if p.parent.name == 'val']
        src = src.split_by_idxs(train_idxs, val_idxs)\
            .label_from_func(lambda x: x, remap=remap, class_mapping=class_mapping)
        device = get_device()
        data = src.databunch(bs=batch_size, device=device, **databunch_kwargs)
        data.meta = meta
        data.remap = remap
        data.classes =  classes
        data.c = data.meta['num_classes']
        data.show_batch = types.MethodType(show_point_cloud_batch_TF, data)
        data.color_mapping = kwargs.get('color_mapping', {k:[random.choice(range(256)) for _ in range(3)]  for k,_ in data.class_mapping.items()})
        data.color_mapping = {int(k):v for k, v in data.color_mapping.items()}
        data.color_mapping = recompute_color_mapping(data.color_mapping, data.classes)
        data.class_mapping = class_mapping
        data.max_point = data.meta['max_point']
        data.extra_dim = data.meta['num_extra_dim']
        data.extra_features = data.meta['extra_features']
        data.block_size = data.meta['block_size']
        ## To accomodate save function to save in correct directory
        data.path = data.path / 'train' 
    else:
        raise Exception("Could not infer dataset type.")

    data.pc_type = dataset_type
    data.path = data.train_ds.path
    ## Below are the lines to make save function work
    data.chip_size = None
    data._image_space_used = None
    data.dataset_type = dataset_type
    data.transform_fn = transform_fn
    return data

def read_xyz_label_from_las(filename_las):
    try_import('laspy')
    import laspy
    msg = 'Loading {}...'.format(filename_las)
    f = laspy.file.File(filename_las, mode='r')    
    h = f.header
    xyzirgb_num = h.point_records_count
    xyz_offset = h.offset
    encoding = h.encoding
    xyz = np.ndarray((xyzirgb_num, 3))
    labels = np.ndarray(xyzirgb_num, np.int16)
    i = 0
    for p in f:
        xyz[i] = [p.x, p.y, p.z]
        labels[i] = p.classification
        i += 1
    f.close()
    return xyz, labels, xyzirgb_num, xyz_offset, encoding

def save_xyz_label_to_las(filename_las, xyz, xyz_offset, encoding, labels):  
    try_import('laspy')
    import laspy
    msg = 'Saving {}...'.format(filename_las)
    h = laspy.header.Header()
    h.dataformat_id = 1
    h.major = 1
    h.minor = 2
    h.min = np.min(xyz, axis=0)
    h.max = np.max(xyz, axis=0)
    h.scale = [1e-3, 1e-3, 1e-3]
    h.offset = xyz_offset
    h.encoding = encoding
    
    f = laspy.file.File(filename_las, mode='w', header=h)    
    for i in range(xyz.shape[0]):
        p = laspy.point.Point()
        p.x = xyz[i,0] / h.scale[0]
        p.y = xyz[i,1] / h.scale[1]
        p.z = xyz[i,2] / h.scale[2]
        p.classification = labels[i]
        p.color = laspy.color.Color()
        p.intensity = 100
        p.return_number = 1
        p.number_of_returns = 1
        p.scan_direction = 1
        p.scan_angle = 0
        f.write(p)
        
    f.close()

def prediction_remap_classes(labels, reclassify_classes, inverse_class_mapping):
    labels = np.vectorize(inverse_class_mapping.get)(labels)
    if reclassify_classes == {}:
        return labels
    else:
        labels = np.vectorize(reclassify_classes.get)(labels)
        return labels

def prediction_selective_classify(labels, las_file, selective_classify):
    all_indexes = list(range(len(labels)))
    classification = las_file.classification
    return np.vectorize(lambda i:labels[i] if labels[i] in selective_classify\
                                        else classification[i])(all_indexes)

def preserved_overwrite(f_out, labels, preserve_classes):
    """
    Does not write class code which is specified
    in preserve_classes param
    """
    orig_classf = f_out.classification
    bool_mat = np.concatenate([orig_classf[:, None] == c for c in preserve_classes], axis=1)
    mask = np.any(bool_mat, axis=1)
    orig_classf[~mask] = labels[~mask]
    return orig_classf

def write_resulting_las(in_las_filename, 
                        out_las_filename, 
                        labels, 
                        num_classes, 
                        data,
                        print_metrics,
                        reclassify_classes={}, 
                        selective_classify=[],
                        preserve_classes=[]):
    try_import('laspy')
    import laspy
    false_positives = [0] * num_classes
    true_positives = [0] * num_classes
    false_negatives = [0] * num_classes
    inverse_class_mapping = {v:k for k,v in data.class_mapping.items()}
    shutil.copy(in_las_filename, out_las_filename)
    f = laspy.file.File(in_las_filename, mode='r')    
    f_out = laspy.file.File(out_las_filename, mode='rw')
    i = 0
    classification = []
    warn_flag = False
    
    ## remap classes
    old_labels = labels.copy()
    labels = prediction_remap_classes(labels, reclassify_classes, inverse_class_mapping)

    if print_metrics:
        for p in f:
            p = f[i]
            current_class = inverse_class_mapping[old_labels[i]] 
            if reclassify_classes != {}:
                current_class = reclassify_classes[current_class]       
            try:
                false_positives[old_labels[i]] += int(p.classification != current_class)
                true_positives[old_labels[i]] += int(p.classification == current_class)
                false_negatives[data.class_mapping[p.classification]] += int(p.classification != current_class)
            except (IndexError, KeyError) as _:
                warn_flag = True

            i += 1

    if selective_classify != []:
        #current_class if current_class in selective_classify else p.classification
        labels = prediction_selective_classify(labels, f, selective_classify)

    if preserve_classes != []:
        labels = preserved_overwrite(f_out, labels, preserve_classes)
            
    f.close()
    f_out.classification = labels.tolist()
    f_out.close()

    # if print_metrics and warn_flag:
    #     logger.warning(f"Some classes in your las file {in_las_filename} do not match the classes the model is trained on")
    #     print_metrics = False
    return false_positives, true_positives, false_negatives

def calculate_metrics(false_positives, true_positives, false_negatives):
    with warnings.catch_warnings():
        warnings.simplefilter("ignore")
        precision = np.divide(true_positives, np.add(true_positives, false_positives))
        recall = np.divide(true_positives, np.add(true_positives, false_negatives))
        f_1 = np.multiply(2.0, np.divide(np.multiply(precision, recall), np.add(precision, recall)))
    return precision, recall, f_1

def get_pred_prefixes(datafolder):
    fs = os.listdir(datafolder)
    preds = []
    for f in fs:
        if f[-8:] == '_pred.h5':
            preds += [f]
    pred_pfx = []
    for p in preds:
        to_check = "_half" #"_zero"
        if to_check in p: 
            pred_pfx += [p.rsplit(to_check, 1)[0]]
    return np.unique(pred_pfx)

def get_predictions(pointcnn_model, data, batch_idx, points_batch, sample_num, batch_size, point_num):
    ## Getting sampling indices
    tile_num = math.ceil((sample_num * batch_size) / point_num)
    indices_shuffle = np.tile(np.arange(point_num), tile_num)[0:sample_num * batch_size]
    np.random.shuffle(indices_shuffle)
    indices_batch_shuffle = np.reshape(indices_shuffle, (batch_size, sample_num, 1))

    model_input = np.concatenate([points_batch[i, s[:, 0]][None] for i, s in enumerate(indices_batch_shuffle)], axis=0)
    
    ## Putting model in evaluation mode and inferencing.
    pointcnn_model.learn.model.eval()
    with torch.no_grad():
        probs = pointcnn_model.learn.model(torch.tensor(model_input).to(pointcnn_model._device).float()).softmax(dim=-1).cpu()
        
    seg_probs = probs.numpy()
    
    probs_2d = np.reshape(seg_probs, (sample_num * batch_size, -1))  ## Complete probs
    predictions = [(-1, 0.0)] * point_num  ## predictions
    
    ## Assigning the confidences and labels to the appropriate index.
    for idx in range(sample_num * batch_size):
        point_idx = indices_shuffle[idx]
        probs = probs_2d[idx, :]
        confidence = np.amax(probs)
        label = np.argmax(probs)
        if confidence > predictions[point_idx][1]:
            predictions[point_idx] = [label, confidence]
        
    return predictions

def inference_las(path,
                  pointcnn_model, 
                  out_path=None, 
                  print_metrics=False, 
                  remap_classes={}, 
                  selective_classify=[],
                  preserve_classes=[]
                  ):
    try_import("h5py")
    import h5py
    import pandas as pd    
    try:
        ## Export data
        path = Path(path)

        if len(list(path.glob('*.las'))) == 0:
            raise Exception(f"The given path({path}) contains no las files.")

        if out_path is None:
            out_path = path / 'results'
        else:    
            out_path = Path(out_path)

        reclassify_classes = remap_classes
        if reclassify_classes != {}:
            if not all([k in pointcnn_model._data.classes for k in reclassify_classes.keys()]):
                raise Exception(f"`remap_classes` dictionary keys are not present in dataset with classes {pointcnn_model._data.classes}.")
            reclassify_classes = {k:reclassify_classes.get(k, k) for k in pointcnn_model._data.class_mapping}

        if selective_classify != []:
            if reclassify_classes != {}:
                values_to_check = np.unique(np.array(list(reclassify_classes.values()))).tolist()
            else:
                values_to_check = list(pointcnn_model._data.classes)

            if not all([k in values_to_check for k in selective_classify]):
                raise Exception(f"`selective_classify` can only contain values from these class values {values_to_check}.")

        prepare_las_data(path.parent,
                        block_size=pointcnn_model._data.block_size[0],
                        max_point_num=pointcnn_model._data.max_point,
                        output_path=path.parent,
                        extra_features=pointcnn_model._data.extra_features,
                        folder_names=[path.stem],
                        segregate=False,
                        print_it=False
        )
        ## Predict and postprocess
        max_point_num = pointcnn_model._data.max_point
        sample_num = pointcnn_model.sample_point_num
        batch_size = 1 * math.ceil(max_point_num / sample_num) 
        filenames = list(glob.glob(str(path/ "*.h5")))

        mb = master_bar(range(len(filenames)))
        for itn in mb:  
            filename = filenames[itn]
            with h5py.File(filename, 'r') as data_h5:
                has_indices = 'indices_split_to_full' in data_h5
                data = data_h5['data'][...].astype(np.float32)  
                data_num =  data_h5['data_num'][...].astype(np.int32)
                indices_split_to_full = data_h5['indices_split_to_full'][...]
            batch_num = data.shape[0]
            labels_pred = np.full((batch_num, max_point_num), -1, dtype=np.int32)
            confidences_pred = np.zeros((batch_num, max_point_num), dtype=np.float32)


            for batch_idx in progress_bar(range(batch_num), parent=mb): 
                points_batch = data[[batch_idx] * batch_size, ...]
                point_num = data_num[batch_idx]
                predictions = get_predictions(pointcnn_model, data, batch_idx, points_batch, sample_num, batch_size, point_num)      
                labels_pred[batch_idx, 0:point_num] = np.array([label for label, _ in predictions])
                confidences_pred[batch_idx, 0:point_num] = np.array([confidence for _, confidence in predictions])

            ## Saving h5 predictions file
            if not os.path.exists(out_path):
                os.makedirs(out_path)
            filename_pred = os.path.join(out_path , Path(filename).stem + '_pred.h5')
            with h5py.File(filename_pred, 'w') as file:
                file.create_dataset('data_num', data=data_num)
                file.create_dataset('label_seg', data=labels_pred)
                file.create_dataset('confidence', data=confidences_pred)                
                if has_indices:
                    file.create_dataset('indices_split_to_full', data=indices_split_to_full)


        ## Merge H5 files and write las files
        SAVE_TO_EXT = '.las'
        LOAD_FROM_EXT = '.las'


        categories_list = get_pred_prefixes(out_path)

        global_false_positives = [0] * pointcnn_model._data.c
        global_true_positives = [0] * pointcnn_model._data.c
        global_false_negatives = [0] * pointcnn_model._data.c

        for category in categories_list:
            output_path = os.path.join(out_path, category + "_pred" + SAVE_TO_EXT)
            if not os.path.exists(os.path.join(out_path)):
                os.makedirs(os.path.join(out_path))
            pred_list = [pred for pred in os.listdir(out_path)
                        if category in pred and pred.rsplit(".", 1)[0].split("_")[-1] == 'pred' and pred[-3:] == '.h5']
            merged_label = None
            merged_confidence = None

            for pred_file in pred_list:

                with h5py.File(os.path.join(out_path, pred_file), mode='r') as data:
                    labels_seg = data['label_seg'][...].astype(np.int64)
                    indices = data['indices_split_to_full'][...].astype(np.int64)
                    confidence = data['confidence'][...].astype(np.float32)
                    data_num = data['data_num'][...].astype(np.int64)

                if merged_label is None:
                    # calculating how many labels need to be there in the output
                    label_length = 0
                    for i in range(indices.shape[0]):
                        label_length = np.max([label_length, np.max(indices[i][:data_num[i]])])
                    label_length += 1
                    merged_label = np.zeros((label_length), dtype=int)
                    merged_confidence = np.zeros((label_length), dtype=float)
                else:
                    label_length2 = 0
                    for i in range(indices.shape[0]):
                        label_length2 = np.max([label_length2, np.max(indices[i][:data_num[i]])])
                    label_length2 += 1
                    if label_length < label_length2:
                        # expanding labels and confidence arrays, as the new file appears having more of them.
                        labels_more = np.zeros((label_length2 - label_length), dtype=merged_label.dtype)
                        conf_more = np.zeros((label_length2 - label_length), dtype=merged_confidence.dtype)
                        merged_label = np.append(merged_label, labels_more)
                        merged_confidence = np.append(merged_confidence, conf_more)
                        label_length = label_length2
                
                for i in range(labels_seg.shape[0]):
                    temp_label = np.zeros((data_num[i]),dtype=int)
                    pred_confidence = confidence[i][:data_num[i]]
                    temp_confidence = merged_confidence[indices[i][:data_num[i]]]

                    temp_label[temp_confidence >= pred_confidence] = merged_label[indices[i][:data_num[i]]][temp_confidence >= pred_confidence]
                    temp_label[pred_confidence > temp_confidence] = labels_seg[i][:data_num[i]][pred_confidence > temp_confidence]

                    merged_confidence[indices[i][:data_num[i]][pred_confidence > temp_confidence]] = pred_confidence[pred_confidence > temp_confidence]
                    merged_label[indices[i][:data_num[i]]] = temp_label

            if len(pred_list) > 0:
                # concatenating source points with the final labels and writing out resulting file
                points_path = os.path.join(path, category + LOAD_FROM_EXT)
                
                false_positives, true_positives, false_negatives = write_resulting_las(points_path,
                                                                                    output_path,
                                                                                    merged_label,
                                                                                    pointcnn_model._data.c,
                                                                                    pointcnn_model._data,
                                                                                    print_metrics,
                                                                                    reclassify_classes,
                                                                                    selective_classify,
                                                                                    preserve_classes)
                global_false_positives = np.add(global_false_positives, false_positives)
                global_true_positives = np.add(global_true_positives, true_positives)
                global_false_negatives = np.add(global_false_negatives, false_negatives)

        if print_metrics:
            index = ['precision', 'recall', 'f1_score']
            inverse_class_mapping = {v:k for k,v in pointcnn_model._data.class_mapping.items()}
            unique_mapped_classes = np.unique(np.array(list(reclassify_classes.values())))
            if len(unique_mapped_classes) == len(pointcnn_model._data.classes) or remap_classes == {}:
                precision, recall, f_1 = calculate_metrics(global_false_positives, global_true_positives, global_false_negatives)
                data = [precision, recall, f_1]
                column_names = [inverse_class_mapping[cval] for cval in range(pointcnn_model._data.c)]
                if reclassify_classes != {}:
                    remapping_class_mapping = {v:reclassify_classes[k] for k,v in pointcnn_model._data.class_mapping.items()}
                    column_names = [remapping_class_mapping[cval] for cval in range(pointcnn_model._data.c)]
                df = pd.DataFrame(data, columns=column_names, index=index)
            else:
                inverse_reclassify_classes = {}
                for k, v in reclassify_classes.items():
                    current_value = inverse_reclassify_classes.get(v, [])
                    current_value.append(k)
                    inverse_reclassify_classes[v] = current_value   
                map_dict = {u: [pointcnn_model._data.class_mapping[k] for k in inverse_reclassify_classes[u]] for u in unique_mapped_classes}
                global_false_positives =  recompute_globals(global_false_positives, map_dict)
                global_true_positives = recompute_globals(global_true_positives, map_dict)
                global_false_negatives = recompute_globals(global_false_negatives, map_dict)
                precision, recall, f_1 = calculate_metrics(global_false_positives, global_true_positives, global_false_negatives)
                data = [precision, recall, f_1]
                column_names = list(map_dict.keys())
                df = pd.DataFrame(data, columns=column_names, index=index)            

            from IPython.display import display
            display(df)
    except KeyboardInterrupt:
        remove_temp_files(path, out_path)
        raise

    remove_temp_files(path, out_path)      

    return out_path

def remove_temp_files(path, out_path):
    for fn in glob.glob(str(path / '*.h5'), recursive=True): ## Remove h5 files in val directory.
        os.remove(fn) 

    for fn in glob.glob(str(out_path / '*.h5'), recursive=True):  ## Remove h5 files in results directory.
        os.remove(fn) 

def recompute_globals(global_count, map_dict):
    return [sum([global_count[ci] for ci in v]) for k,v in map_dict.items()]

def raise_maxpoint_warning(idx_file, kwargs, logger, max_display_point, save_html=False):
    if not save_html:
        if idx_file == 0:
            if 'max_display_point' not in kwargs.keys():
                logger.warning(f"Randomly sampling {max_display_point} points for visualization. You can adjust this using the `max_display_point` parameter.")

def get_title_text(idx, save_html, max_display_point):
    title_text = 'Ground Truth / Predictions' if idx==0 else ''
    if save_html:
        title_text = f'Ground Truth / Predictions (Displaying randomly sampled {max_display_point} points.)' if idx==0 else ''
    return title_text

def inverse_remap_predictions(predictions, class_mapping):
    if isinstance(predictions, torch.Tensor):
        remapped_predictions = torch.zeros_like(predictions)
    else:
        remapped_predictions = np.zeros_like(predictions).astype(int)
    for k,v in class_mapping.items():
        remapped_predictions[predictions == v] = k
    return remapped_predictions    

def show_results(self, rows, color_mapping=None, **kwargs):

    """
    It will plot results from your trained model with ground truth on the
    left and predictions on the right.

    =====================   ===========================================
    **Argument**            **Description**
    ---------------------   -------------------------------------------
    rows                    Optional rows. Number of rows to show. Deafults
                            value is 2.
    ---------------------   -------------------------------------------
    color_mapping           Optional dictionary. Mapping from class value
                            to RGB values. Default value example:
                            {0:[220,220,220], 2:[255,0,0], 6:[0,255,0]}.
    =====================   ===========================================

    **kwargs**

    =====================   ===========================================
    **Argument**            **Description**
    ---------------------   -------------------------------------------
    mask_class              Optional array of integers. Array containing
                            class values to mask. Default value is [].    
    ---------------------   -------------------------------------------
    width                   Optional integer. Width of the plot. Default 
                            value is 750.
    ---------------------   -------------------------------------------
    height                  Optional integer. Height of the plot. Default
                            value is 512
    ---------------------   -------------------------------------------
    max_display_point       Optional integer. Maximum number of points
                            to display. Default is 20000.
    ---------------------   -------------------------------------------
    return_fig              Optional bool. Flag set as True when the matplotlib
                            figure is needed as a return object. Used for TB integration
    ---------------------   -------------------------------------------
    =====================   ===========================================
    """
    
    filter_outliers = False
    try_import("h5py")
    try_import('plotly')
    import h5py    
    import plotly
    import plotly.graph_objects as go
    from plotly.subplots import make_subplots    
    import random
    mask_class = kwargs.get('mask_class', [])    
    save_html = kwargs.get('save_html', False)
    save_path = kwargs.get('save_path', '.')
    return_fig = kwargs.get('return_fig', False)
    max_display_point = get_max_display_points(self._data, kwargs)
    rows = min(rows, self._data.batch_size)
    color_mapping = self._data.color_mapping if color_mapping is None else color_mapping
    color_mapping = recompute_color_mapping(color_mapping, self._data.classes)       
    color_mapping = np.array(list(color_mapping.values())) / 255
    if save_html:
        max_display_point = 20000

    idx = 0
    keys = list(self._data.meta['files'].keys()).copy()
    keys = [f for f in keys if Path(f).parent.stem == 'val']
    random.shuffle(keys)

    for idx_file, fn in enumerate(keys):        
        fig = make_subplots(rows=1, cols=2, specs=[[{'type': 'scene'}, {'type': 'scene'}]])        
        num_files = self._data.meta['files'][fn]['idxs']
        block_center = self._data.meta['files'][fn]['block_center']
        block_center = np.array(block_center)
        block_center[0][2], block_center[0][1] = block_center[0][1], block_center[0][2]
        if num_files == []:
            continue
        if not Path(fn).is_absolute():
            fn = str(self._data.path / fn)            
        idxs = [h5py.File(fn[:-3] + f'_{i}.h5', 'r') for i in num_files]
        pc = []
        labels = []
        pred_class = []
        pred_confidence = []
        for i in idxs:
            # print(f'Running Show Results: Processing {nn+1} of {len(idxs)} blocks.', end='\r')
            current_block = i['unnormalized_data'][:, :3]
            data_num = i['data_num'][()] 
            data = i['data'][:] 
            pc.append(current_block[:data_num])
            labels.append(i['label_seg'][:data_num])

            max_point_num = self._data.max_point
            sample_num = self.sample_point_num
            batch_size = 1 * math.ceil(max_point_num / sample_num) 
            data = data[None]
            batch_idx = 0
            points_batch = data[[batch_idx] * batch_size, ...]
            point_num = data_num
            predictions = np.array(get_predictions(self, data, batch_idx, points_batch, sample_num, batch_size, point_num))
            pred_class.append(predictions[:, 0])
            pred_confidence.append(predictions[:, 1])
            i.close()
            
        if pc == []:
            continue         
                
        pc = np.concatenate(pc, axis=0)
        labels = np.concatenate(labels, axis=0)
        unmapped_labels = labels.copy().astype(int)
        if self._data.remap:
            labels = remap_labels(labels, self._data.class_mapping)
        pred_class = np.concatenate(pred_class, axis=0).astype(int)
        unmapped_pred_class = inverse_remap_predictions(pred_class, self._data.class_mapping)
        sample_idxs = mask_classes(labels, mask_class, self._data.class_mapping if self._data.remap else None)
        sampled_pc = pc[sample_idxs]
        if sampled_pc.shape[0] == 0:
            continue
        x, y, z = recenter(sampled_pc).transpose(1,0)       
        if filter_outliers:
            ## Filter on the basis of std.
            mask = filter_pc(pc)
        else:
            ## all points
            mask = x != None

        if sample_idxs.sum() > max_display_point:
            raise_maxpoint_warning(idx_file, kwargs, logger, max_display_point, save_html)
            mask = np.random.randint(0, sample_idxs.sum(), max_display_point)              
        
        color_list_true =  color_mapping[labels[sample_idxs]][mask].tolist()
        color_list_pred = color_mapping[pred_class[sample_idxs]][mask].tolist()
        
        scene=dict(aspectmode='data')


        fig.add_trace(go.Scatter3d(x=x[mask], y=z[mask], z=y[mask], 
                                        mode='markers', marker=dict(size=1, color=color_list_true),
                                        text=class_string(unmapped_labels[sample_idxs][mask])), row=1, col=1)

        fig.add_trace(go.Scatter3d(x=x[mask], y=z[mask], z=y[mask], 
                                        mode='markers', marker=dict(size=1, color=color_list_pred),
                                        text=class_string(unmapped_pred_class[sample_idxs][mask], prefix='pred_')), row=1, col=2)

        title_text = get_title_text(idx, save_html, max_display_point)

        fig.update_layout(
            scene=scene,
            scene2=scene,
            title_text=title_text,
            width=kwargs.get('width', 750),
            height=kwargs.get('width', 512),
            showlegend=False,
            title_x=0.5
        )

        if save_html:
            save_path = Path(save_path)
            plotly.io.write_html(fig, str(save_path / 'show_results.html'))
            # remove plotly orca dep
            # fig.write_image(str(save_path / 'show_results.png'))
            return
        else:
            fig.show()

        if idx == rows-1:
            break
        idx += 1
    if return_fig:
        return fig

def compute_precision_recall(self):
    from ..models._pointcnn_utils import get_indices
    import pandas as pd 

    valid_dl = self._data.valid_dl
    model = self.learn.model.eval()

    false_positives = [0] * self._data.c
    true_positives = [0] * self._data.c
    false_negatives = [0] * self._data.c
    class_count = [0] * self._data.c

    all_y = []
    all_pred = []
    for x_in, y_in in iter(valid_dl):
        x_in, point_nums = x_in   ## (batch, total_points, num_features), (batch,)
        batch, _, num_features = x_in.shape
        indices = torch.tensor(get_indices(batch, self.sample_point_num, point_nums.long())).to(x_in.device)
        indices = indices.view(-1, 2).long()
        x_in = x_in[indices[:, 0], indices[:, 1]].view(batch, self.sample_point_num, num_features).contiguous()  ## batch, self.sample_point_num, num_features                
        y_in = y_in[indices[:, 0], indices[:, 1]].view(batch, self.sample_point_num).contiguous().cpu().numpy() ## batch, self.sample_point_num        
        with torch.no_grad():
            preds = model(x_in).detach().cpu().numpy()
        predicted_labels = preds.argmax(axis=-1)
        all_y.append(y_in.reshape(-1))
        all_pred.append(predicted_labels.reshape(-1))

    all_y = np.concatenate(all_y)
    all_pred = np.concatenate(all_pred)
    
    for i in range(len(all_y)): 
        class_count[all_y[i]] += 1       
        false_positives[all_pred[i]] += int(all_y[i] != all_pred[i])
        true_positives[all_pred[i]] += int(all_y[i] == all_pred[i])
        false_negatives[all_y[i]] += int(all_y[i] != all_pred[i])        
    
    
    precision, recall, f_1 = calculate_metrics(false_positives, true_positives, false_negatives)
    data = [precision, recall, f_1]
    index = ['precision', 'recall', 'f1_score']
    inverse_class_mapping = {v:k for k,v in self._data.class_mapping.items()} 
    df = pd.DataFrame(data, columns=[inverse_class_mapping[cval] for cval in range(self._data.c)], index=index) 
    return df

def gauss_clip(mu, sigma, clip):
    v = random.gauss(mu, sigma)
    v = max(min(v, mu + clip * sigma), mu - clip * sigma)
    return v


def uniform(bound):
    return bound * (2 * random.random() - 1)


def scaling_factor(scaling_param, method):
    try:
        scaling_list = list(scaling_param)
        return random.choice(scaling_list)
    except:
        if method == 'g':
            return gauss_clip(1.0, scaling_param, 3)
        elif method == 'u':
            return 1.0 + uniform(scaling_param)


def rotation_angle(rotation_param, method):
    try:
        rotation_list = list(rotation_param)
        return random.choice(rotation_list)
    except:
        if method == 'g':
            return gauss_clip(0.0, rotation_param, 3)
        elif method == 'u':
            return uniform(rotation_param)


def get_xforms(xform_num, rotation_range=(0, 0, 0, 'u'), scaling_range=(0.0, 0.0, 0.0, 'u'), order='XYZ'):
    xforms = np.empty(shape=(xform_num, 3, 3))
    rotations = np.empty(shape=(xform_num, 3, 3))
    for i in range(xform_num):
        rx = rotation_angle(rotation_range[0], rotation_range[3])
        ry = rotation_angle(rotation_range[1], rotation_range[3])
        rz = rotation_angle(rotation_range[2], rotation_range[3])
        rotation = R.from_euler(order, [rx, ry, rz]).as_matrix()

        sx = scaling_factor(scaling_range[0], scaling_range[3])
        sy = scaling_factor(scaling_range[1], scaling_range[3])
        sz = scaling_factor(scaling_range[2], scaling_range[3])
        scaling = np.diag([sx, sy, sz])

        xforms[i, :] = scaling * rotation
        rotations[i, :] = rotation
    return xforms, rotations

def augment(points, xforms, range=None):
    points_xformed = points@xforms
    if range is None:
        return points_xformed
    
    jitter_data = range * points.new(np.random.randn(*points_xformed.shape))
    jitter_clipped = torch.clamp(jitter_data, -5 * range, 5 * range)
    return points_xformed + jitter_clipped

class Transform3d(object):

    """
    Creates a 3D transformation that can be used in `prepare_data` 
    to apply data augmentation to blocks, with a 50 % probability.

    =====================   ===========================================
    **Argument**            **Description**
    ---------------------   -------------------------------------------
    rotation_range          Optional tuple of length 4. It contains a list
                            of angles(in radians) for X, Z and Y coordinates
                            respectively. These angles will rotate the point
                            cloud block according to the randomly selected angle.
                            The fourth value in the tuple is the sampling method
                            where 'u' means uniform and 'g' means gaussian.
                            Intrinsic rotation will take place.
                            Default: [math.pi / 72, math.pi, math.pi / 72, 'u'].
    ---------------------   -------------------------------------------
    scaling_range           Optional tuple of length 4. It contains a list
                            of scaling ranges[0-1] which will scale the points.
                            Please keep it a very small number otherwise,
                            point cloud block may get distorted. The fourth
                            value in the tuple is the sampling method
                            where 'u' means uniform and 'g' means gaussian.
                            Default: [0.05, 0.05, 0.05, 'g'].
    ---------------------   -------------------------------------------
    jitter                  Optional float. The scale to which randomly
                            jitter the points in the point cloud block.
                            Default: 0.0.
    =====================   ===========================================
    
    :returns: `Transform3d` object
    """

    def __init__(self, rotation_range=[math.pi / 72, math.pi, math.pi / 72, 'u'],
                 scaling_range=[0.05, 0.05, 0.05, 'g'], 
                 jitter=0.):
        self.rotation_range = rotation_range
        self.scaling_range = scaling_range
        self.order = 'XYZ'
        self.jitter = jitter

    def __call__(self, x_in):
        xforms, _ = get_xforms(x_in.shape[0], rotation_range=self.rotation_range, scaling_range=self.scaling_range, order=self.order)
        return augment(x_in[:, :, :3], x_in.new(xforms), x_in.new(np.array(self.jitter)))

def save_h5(filename, labels_pred, confidences_pred):
    try_import('h5py')
    import h5py
    filename = Path(filename)
    if not filename.parent.exists():
        filename.parent.mkdir(parents=True, exist_ok=True)

    filename_pred = filename.parent / (filename.stem + '_pred.h5')
    file = h5py.File(filename_pred, 'w')
    file.create_dataset('label_seg', data=labels_pred)
    file.create_dataset('confidence', data=confidences_pred)
    file.close()


def predict_h5(self, path, output_path, print_metrics=False):
    """
    self: PointCNN object
    path: path/to/h5/files/exported/by/tool
    """
    path = Path(path)
    if output_path is None:
        output_path = path / 'results'
    else:
        output_path = Path(output_path)

    point_cloud_dataset = PointCloudDataset(path, max_point=self._data.max_point, extra_dim=self._data.extra_dim, class_mapping=self._data.class_mapping)
    batch_size = 1 * math.ceil(self._data.max_point / self.sample_point_num)   
    max_point_num = self._data.max_point

    folder_name = Path(point_cloud_dataset.relative_files[0].decode()).parts[0]
    name_txt_path = path / folder_name / 'Name.txt'
    (output_path / folder_name).mkdir(parents=True, exist_ok=True)
    shutil.copy(str(name_txt_path), str(output_path / folder_name))

    current_file_name = ''
    for i in progress_bar(range(len(point_cloud_dataset))):
        tile = point_cloud_dataset.tiles[i]
        h5_file = point_cloud_dataset.h5files[tile[0]]
        fname = point_cloud_dataset.filenames[tile[0]]

        if fname != current_file_name:
            low = 0
            if i!=0:
                save_h5(output_path / point_cloud_dataset.relative_files[int(tile[0] - 1)].decode(), labels_pred, confidences_pred)
            current_file_name = fname
            batch_num, _ = h5_file['xyz'].shape
            labels_pred = np.full(batch_num, -1, dtype=np.int8)
            confidences_pred = np.zeros(batch_num, dtype=np.float32)

        (data, point_num), classification = point_cloud_dataset[i]
        data = data[None]
        points_batch = data[[0] * batch_size]
        predictions = get_predictions(self, data, 0, points_batch, self.sample_point_num, batch_size, point_num)
        high = low + point_num
        labels_pred[low:high] = np.array([label for label, _ in predictions])
        confidences_pred[low:high] = np.array([confidence for _, confidence in predictions])
        low = high

        if i == (len(point_cloud_dataset) - 1):
            save_h5(output_path / point_cloud_dataset.relative_files[tile[0]].decode(), labels_pred, confidences_pred)

        if print_metrics:
            if classification is None and i == 0:
                logger.warning("classification codes are not present in the h5 file.")
            else:
                pass        


    return output_path

def calculate_per_class_stats(all_pred, all_y, total_classes):

    true_positives = [0] * total_classes 
    false_positives = [0] * total_classes
    false_negatives = [0] * total_classes
    class_count = [0] * total_classes

    for i in range(len(all_y)): 
        class_count[all_y[i]] += 1       
        false_positives[all_pred[i]] += int(all_y[i] != all_pred[i])
        true_positives[all_pred[i]] += int(all_y[i] == all_pred[i])
        false_negatives[all_y[i]] += int(all_y[i] != all_pred[i])       

    return true_positives, false_positives, false_negatives

def show_results_tool(self, rows, color_mapping=None, **kwargs):

    """
    It will plot results from your trained model with ground truth on the
    left and predictions on the right.

    =====================   ===========================================
    **Argument**            **Description**
    ---------------------   -------------------------------------------
    rows                    Optional rows. Number of rows to show. Deafults
                            value is 2.
    ---------------------   -------------------------------------------
    color_mapping           Optional dictionary. Mapping from class value
                            to RGB values. Default value example:
                            {0:[220,220,220], 2:[255,0,0], 6:[0,255,0]}.
    =====================   ===========================================

    **kwargs**

    =====================   ===========================================
    **Argument**            **Description**
    ---------------------   -------------------------------------------
    mask_class              Optional array of integers. Array containing
                            class values to mask. Default value is [0].    
    ---------------------   -------------------------------------------
    width                   Optional integer. Width of the plot. Default 
                            value is 750.
    ---------------------   -------------------------------------------
    height                  Optional integer. Height of the plot. Default
                            value is 512
    ---------------------   -------------------------------------------
    max_display_point       Optional integer. Maximum number of points
                            to display. Default is 20000.                               
    =====================   ===========================================
    """
    filter_outliers = False
    try_import("h5py")
    try_import('plotly')
    import h5py    
    import plotly
    import plotly.graph_objects as go
    from plotly.subplots import make_subplots    
    import random
    mask_class = kwargs.get('mask_class', [])    
    save_html = kwargs.get('save_html', False)
    save_path = kwargs.get('save_path', '.')
    max_display_point = get_max_display_points(self._data, kwargs)
    data = self._data
    rows = min(rows, data.batch_size)
    color_mapping = data.color_mapping if color_mapping is None else color_mapping
    color_mapping = recompute_color_mapping(color_mapping, self._data.classes)       
    color_mapping = np.array(list(color_mapping.values())) / 255    
    if save_html:
        max_display_point = 20000
    ## dataset tiles Get all files from the tiles
    tile_file_indices = self._data.train_ds.tiles[:, 0]
    ## iterate: on files
    for idx, file_idx in enumerate(data.val_files):
        ## Create subplot
        fig = make_subplots(rows=1, cols=2, specs=[[{'type': 'scene'}, {'type': 'scene'}]])

        # read that file
        indices = (tile_file_indices == file_idx).nonzero()[0]  

        ## predict on each block by iterating.
        pred_batch_size = 1 * math.ceil(self._data.max_point / self.sample_point_num) 
        labels = []
        pc = []
        pred_class = []
        for block_idx in indices:
            (block, point_num), classification = data.train_ds[block_idx]
            block = block[None]
            points_batch = block[[0] * 1]
            predictions = np.array(get_predictions(self, block, 0, points_batch, self.sample_point_num, pred_batch_size, point_num))
            pred_class.append(predictions[:point_num, 0])
            pc.append(block[0, :point_num].cpu().numpy())
            labels.append(classification[:point_num].cpu().numpy())

        labels = np.concatenate(labels)
        pc = np.concatenate(pc)
        pred_class = np.concatenate(pred_class, axis=0)

        unmapped_labels = labels.copy().astype(int)
        unmapped_predictions = inverse_remap_predictions(pred_class, self._data.class_mapping)
        ## remapping the labels from 0-N
        if self._data.remap:
            labels = remap_labels(labels, self._data.class_mapping)

        ## sample points
        sample_idxs = mask_classes(labels, mask_class, self._data.class_mapping if self._data.remap else None)
        sampled_pc = pc[sample_idxs]
        if sampled_pc.shape[0] == 0:
            continue  
        sampled_pc = sampled_pc[:, :3]
        x, y, z = recenter(sampled_pc).transpose(1,0)   

        ## resample points if exeeds limits.
        if sample_idxs.sum() > max_display_point:
            raise_maxpoint_warning(idx, kwargs, logger, max_display_point, save_html)
            mask = np.random.randint(0, sample_idxs.sum(), max_display_point) 
        
        ## Apply cmap
        color_list_true =  color_mapping[labels[sample_idxs]][mask].tolist()
        color_list_pred = color_mapping[pred_class[sample_idxs].astype(int)][mask].tolist()        

        ## Plot
        scene=dict(aspectmode='data')


        fig.add_trace(go.Scatter3d(x=x[mask], y=y[mask], z=z[mask], 
                                        mode='markers', marker=dict(size=1, color=color_list_true),
                                        text=class_string(unmapped_labels[sample_idxs][mask])), row=1, col=1)

        fig.add_trace(go.Scatter3d(x=x[mask], y=y[mask], z=z[mask], 
                                        mode='markers', marker=dict(size=1, color=color_list_pred),
                                        text=class_string(unmapped_predictions[sample_idxs][mask], prefix='pred_')), row=1, col=2)


        title_text = get_title_text(idx, save_html, max_display_point)
        fig.update_layout(
            scene=scene,
            scene2=scene,
            title_text=title_text,
            width=kwargs.get('width', 750),
            height=kwargs.get('width', 512),
            showlegend=False,
            title_x=0.5
        )

        if save_html:
            save_path = Path(save_path)
            plotly.io.write_html(fig, str(save_path / 'show_results.html'))
            # remove plotly orca dep
            # fig.write_image(str(save_path / 'show_results.png'))
            return
        else:
            fig.show()

        if idx == rows-1:
            break



import arcgis


class Header(object):
    """
    Creates a dashboard header widget.

    =========================   ===========================================
    **Argument**                **Description**
    -------------------------   -------------------------------------------
    title                       Optional string. Title of the header.
    -------------------------   -------------------------------------------
    subtitle                    Optional string. Subtitle of the header.
    -------------------------   -------------------------------------------
    margin                      Optional boolean. Set True to add margin to
                                header position.
    -------------------------   -------------------------------------------
    size                        Optional string. Define size of header from
                                small, medium, large.
                                Default is medium.
    -------------------------   -------------------------------------------
    logo_image_url              Optional url string. Define a logo image.
    -------------------------   -------------------------------------------
    logo_url                    Optional url string. Define a hyperlink for
                                the logo image.
    -------------------------   -------------------------------------------
    background_image_url        Optional url string. Add a background image
                                to the header.
    -------------------------   -------------------------------------------
    background_image_size       Optional string. Select size of the image.
                                Options:
                                    fit-width
                                    fit-height
                                    fit-both
                                    repeat
    -------------------------   -------------------------------------------
    background_image_position   Optional string. Define the image position
                                when using fit-width or fit-height.
                                Allowed options:
                                fit-height: left, center, right
                                fit-width: top, middle, bottom
    -------------------------   -------------------------------------------
    signout_link                Optional boolean. Parameter to save the best model
                                during training. If set to `True` the best model
                                based on validation loss will be saved during
                                training.
    -------------------------   -------------------------------------------
    menu_links                  Optional list of tuples. Each tuple contains
                                string label and string url.
    =========================   ===========================================
    """
    def __init__(
            self,
            title=None,
            subtitle=None,
            margin=True,
            size="medium",
            logo_image_url=None,
            logo_url=None,
            background_image_url=None,
            background_image_size="fit-width",
            background_image_position="left",
            signout_link=False,
            menu_links=None
    ):
        # Header starts here.

        self._size = "medium"
        self._title = None
        self._subtitle = None
        self._margin = False
        self.type = "headerPanel"

        self.size = size
        self.title = title
        self.subtitle = subtitle
        self.margin = margin

        self._logo_image_url = logo_image_url
        self._logo_url = logo_url

        self._background_image = background_image_url
        self._background_image_placement = ("fit-height", "left")
        self.background_image_placement = (background_image_size, background_image_position)

        self._sign_out_link = signout_link
        self._menu_links = []

        self.menu_links = menu_links

        self._text_color = None
        self._background_color = None
        self._selectors = []
        # Header ends here.

    def _repr_html_(self):
        from arcgis.apps.dashboard import Dashboard
        url = Dashboard._publish_random(self)
        return f"""<iframe src={url} width=300 height=300>"""

    @property
    def size(self):
        """
        Return size of the header. small, medium or large
        :return: string
        """
        return self._size

    @size.setter
    def size(self, value):
        """
        Set size of the header. Value can be small, medium or large.
        """
        self._size = value
        if value not in ["small", "medium", "large"]:
            self._size = "medium"

    @property
    def title(self):
        """
        :return: Returns title of the header
        """
        return self._title

    @title.setter
    def title(self, value):
        """
        Set title of the header. A string.
        """
        if isinstance(value, str):
            self._title = value

    @property
    def subtitle(self):
        """
        :return: Subtitle of the header. A string.
        """
        return self._subtitle

    @subtitle.setter
    def subtitle(self, value):
        """
        Set subtitle of the header.
        """
        self._subtitle = value
        if not isinstance(value, str):
            self._subtitle = None

    @property
    def margin(self):
        """
        :return: Header margin, True or False
        """
        return self._margin

    @margin.setter
    def margin(self, value):
        """
        Set margin True or False.
        """
        self._margin = value
        if not isinstance(value, bool):
            self._margin = True

    @property
    def logo_image_url(self):
        """
        :return: Logo Image Url
        """
        return self._logo_image_url

    @logo_image_url.setter
    def logo_image_url(self, value):
        """
        :return: Set Logo image url.
        """
        self._logo_image_url = value

    @property
    def logo_url(self):
        """
        :return: Logo image hyperlink
        """
        return self._logo_url

    @logo_url.setter
    def logo_url(self, value):
        """
        Set Logo image hyperlink
        """
        self._logo_url = value

    @property
    def background_image_url(self):
        """
        :return: Background image url.
        """
        return self._background_image

    @background_image_url.setter
    def background_image_url(self, value):
        """
        Set background image url.
        """
        self._background_image = value

    @property
    def background_image_placement(self):
        """
        :return: Background image position.
        If fit-height then left, right, center
        If fit-width then top, bottom, middle
        fit-both, repeat then None
        """
        return self._background_image_placement

    @background_image_placement.setter
    def background_image_placement(self, value):
        sizing = "fit-height"
        position = "left"

        size_position_mapping = {
            "fit-height": ["left", "center", "right"],
            "fit-width": ["top", "middle", "bottom"],
            "fit-both": [],
            "repeat": []
        }

        temp_position = None
        if isinstance(value, tuple):
            if size_position_mapping.get(value[0]):
                sizing = value[0]
                temp_position = value[1]
        elif size_position_mapping.get(value):
            sizing = value

        if temp_position:
            if len(size_position_mapping.get(sizing)) == 0:
                position = None
            elif temp_position in size_position_mapping.get(sizing):
                position = temp_position
            else:
                position = size_position_mapping.get(sizing)[0]

        self._background_image_placement = (sizing, position)

    @property
    def sign_out_link(self):
        """
        :return: Sign out link enabled or disabled.
        """
        return self.sign_out_link

    @sign_out_link.setter
    def sign_out_link(self, value):
        """
        Enable or disable the signout link.
        """
        self._sign_out_link = value
        if not isinstance(value, bool):
            self._sign_out_link = True

    @property
    def menu_links(self):
        """
        :return: List of menu links.
        """
        return self._menu_links

    @menu_links.setter
    def menu_links(self, value):
        """
        Set list of tuples of menu links.
        For example: [("label", "url")]
        """
        if isinstance(value, list):
            for val in value:
                if isinstance(val, tuple) and len(val) == 2:
                    self._menu_links.append(val)
        elif isinstance(value, tuple) and len(value) == 2:
            self._menu_links.append(value)

    @property
    def text_color(self):
        """
        :return: None for default text color else HEX code.
        """
        return self._text_color

    @text_color.setter
    def text_color(self, value):
        """
        Set HEX code for text color.
        """
        self._text_color = value
        if not isinstance(value, str):
            self._text_color = None

    @property
    def background_color(self):
        """
        :return: Background color HEX code.
        """
        return self._background_color

    @background_color.setter
    def background_color(self, value):
        """
        Set background color HEX code.
        """
        self._background_color = value
        if not isinstance(value, str):
            self._background_color = None

    def add_selector(self, selector):
        """
        Add Number Selector, Category Selector or Date Picker widget.
        """
        self._selectors.append(selector)

    def _convert_to_json(self):
        header_panel = {
            'type': 'headerPanel',
            'size': self.size,
            'showSignOutMenu': self._sign_out_link,
            'menuLinks': self._menu_links,
            'showMargin': self.margin,
            'selectors': [selector._convert_to_json() for selector in self._selectors]
        }

        if self.title:
            header_panel['title'] = self.title

        if self.subtitle:
            header_panel['subtitle'] = self.subtitle

        if self._background_image:
            header_panel['backgroundImageUrl'] = self._background_image
            header_panel['backgroundImageSizing'] = self._background_image_placement[0]
            header_panel['normalBackgroundImagePlacement'] = self._background_image_placement[1]
            header_panel['horizontalBackgroundImagePlacement'] = self._background_image_placement[1]

        if self._logo_image_url:
            header_panel['logoImageURL'] = self._logo_image_url
            header_panel['logoURL'] = self._logo_url

        if self._text_color:
            header_panel['titleTextColor'] = self._text_color

        if self._background_color:
            header_panel['backgroundColor'] = self._background_color

        return header_panel



import json

class RasterInfo(object):
    """
    Class allows to create RasterInfo object that describes a set of raster properties to
    facilitate the creation of local raster dataset using the Raster class (requires arcpy)

    A RasterInfo object can be created by instantiating it from a dictionary,
    or by calling an Imagery Layer/Raster object's raster_info property

    Information about the raster can also be set through the following properties available on the RasterInfo object:
    band_count, extent , pixel_size_x, pixel_size_y, pixel_type, block_height, block_width, no_data_values, spatial_reference

    To construct RasterInfo object from a dictionary, use from_dict method on this class.

    .. code-block:: python

        # Usage Example 1: This example creates a new Raster object from the raster_info of another Raster object. (requires arcpy)
        raster_obj = Raster(<raster dataset path>)
        ras_info = RasterInfo(raster_obj.raster_info)
        rinfo_based_ras = Raster(rasInfo2)

        #To write pixel values to this temporary Raster object:
        rinfo_based_ras.write(<numpy_array>)

        #To save this temporary raster locally:
        rinfo_based_ras.save(r"C:\data\persisted_raster.crf")

    RasterInfo object can also be used in raster functions that take in raster info as a parameter. (does not require arcpy)
    example: As value to the raster_info parameter for arcgis.raster.functions.constant_raster() and arcgis.raster.functions.random_raster()

    """
    def __init__(self, raster_info_dict=None):
        self._band_count = None
        self._extent = None
        self._pixel_size_x = None
        self._pixel_size_y = None
        self._pixel_type = None
        self._block_height = None
        self._block_width = None
        self._spatial_reference = None
        self._no_data_values = None
        self._dict=raster_info_dict


    def __repr__(self):
        """returns object as string"""
        return json.dumps(self.to_dict())

    __str__ = __repr__
    
    @property
    def band_count(self):
        """
        Information about the band count of a raster.
        """
        return self._band_count

    @band_count.setter
    def band_count(self, value):
        self._band_count = value

    @property
    def extent(self):
        """
        Information about the extent of a raster.
        """
        return self._extent

    @extent.setter
    def extent(self, value):
        self._extent = value

    @property
    def pixel_type(self):
        """
        Information about the pixel type of a raster.
        """
        return self._pixel_type

    @pixel_type.setter
    def pixel_type(self, value):
        self._pixel_type = value

    @property
    def pixel_size_x(self):
        """
        Information about the pixel size of a raster in x direction 
        """
        return self._pixel_size_x

    @pixel_size_x.setter
    def pixel_size_x(self, value):
        self._pixel_size_x = value

    @property
    def pixel_size_y(self):
        """
        Information about the pixel size of a raster in y direction 
        """
        return self._pixel_size_y

    @pixel_size_y.setter
    def pixel_size_y(self, value):
        self._pixel_size_y = value


    @property
    def block_height(self):
        """
        Information about the block height
        """
        return self._block_height

    @block_height.setter
    def block_height(self, value):
        self._block_height = value

    @property
    def block_width(self):
        """
        Information about the block width
        """
        return self._block_width

    @block_width.setter
    def block_width(self, value):
        self._block_width = value

    @property
    def no_data_values(self):
        """
        Information about the no_data_values
        """
        return self._no_data_values

    @no_data_values.setter
    def no_data_values(self, value):
        self._no_data_values = value

    @property
    def spatial_reference(self):
        """
        Information about the spatial reference of a raster.
        """
        return self._spatial_reference

    @extent.setter
    def spatial_reference(self, value):
        self._spatial_reference = value



    def to_dict(self):
        """
        To return Raster Info in dictionary format
        """
        #rinfo_dict = self.__dict__
        new_rinfo_dict = {}
        if self._dict is None:
            if self.band_count is not None:
                new_rinfo_dict.update({"bandCount":self.band_count})
            if self.extent is not None:
                new_rinfo_dict.update({"extent":self.extent})
            if self.pixel_size_x is not None:
                new_rinfo_dict.update({"pixelSizeX":self.pixel_size_x})
            if self.pixel_size_y is not None:
                new_rinfo_dict.update({"pixelSizeY":self.pixel_size_y})
            if self.pixel_type is not None:
                new_rinfo_dict.update({"pixelType":self.pixel_type})
            if self.block_height is not None:
                new_rinfo_dict.update({"blockHeight":self.block_height})
            if self.block_width is not None:
                new_rinfo_dict.update({"blockWidth":self.block_width})
            if self.spatial_reference is not None:
                new_rinfo_dict.update({"spatialReference":self.spatial_reference})
            if self.no_data_values is not None:
                new_rinfo_dict.update({"noDataValues":self.no_data_values})
        else:
            new_rinfo_dict=self._dict
        return new_rinfo_dict

    def from_dict(self, raster_info_dict):
        """
        Function used to initialise RasterInfo object from raster info dictionary

        .. code-block:: python

            # Usage Example : 
            rinfo = RasterInfo()
            rinfo.from_dict({'bandCount': 3,
                             'extent': {"xmin": 4488761.95,
                                         "ymin": 5478609.805,
                                         "xmax": 4489727.05,
                                         "ymax": 5479555.305,
                                         "spatialReference": {
                                           "wkt": "PROJCS[\"Deutsches_Hauptdreiecksnetz_Transverse_Mercator\",
                                           GEOGCS[\"GCS_Deutsches_Hauptdreiecksnetz\",DATUM[\"D_Deutsches_Hauptdreiecksnetz\",
                                           SPHEROID[\"Bessel_1841\",6377397.155,299.1528128]],PRIMEM[\"Greenwich\",0.0],
                                           UNIT[\"Degree\",0.0174532925199433]],PROJECTION[\"Transverse_Mercator\"],
                                           PARAMETER[\"false_easting\",4500000.0],PARAMETER[\"false_northing\",0.0],
                                           PARAMETER[\"central_meridian\",12.0],PARAMETER[\"scale_factor\",1.0],
                                           PARAMETER[\"latitude_of_origin\",0.0],UNIT[\"Meter\",1.0]]"
                                         }}, 
                             'pixelSizeX': 0.0999999999999614, 
                             'pixelSizeY': 0.1, 
                             'pixelType': 'U8'})
        """
        if raster_info_dict is not None and isinstance(raster_info_dict, dict):

            if "extent" in raster_info_dict:
                self.extent = raster_info_dict["extent"]

            if "bandCount" in raster_info_dict:
                self.band_count = raster_info_dict["bandCount"]

            if "pixelType" in raster_info_dict:
                self.pixel_type = raster_info_dict["pixelType"]

            if "pixelSizeX" in raster_info_dict:
                self.pixel_size_x = raster_info_dict["pixelSizeX"]

            if "pixelSizeY" in raster_info_dict:
                self.pixel_size_y = raster_info_dict["pixelSizeY"]

            self._dict = raster_info_dict



import uuid
import arcgis
from .._utils._basewidget import _BaseWidget


class MapLegend(_BaseWidget):
    """
    Create a MapLegend widget for Dashboard

    =========================   ===========================================
    **Argument**                **Description**
    -------------------------   -------------------------------------------
    map_widget                  Required web map widget. Legend for this
                                Map widget is displayed.
                                This map widget needs to be a part of the
                                final Dashboard as well.
    -------------------------   -------------------------------------------
    name                        Optional String. Name of the widget.
    -------------------------   -------------------------------------------
    title                       Optional string. Title of the widget.
    -------------------------   -------------------------------------------
    description                 Optional string. Description of the widget.
    =========================   ===========================================
    """
    def __init__(self, map_widget, name='MapLegend', title="", description=""):
        super().__init__(name, title, description)

        self._map_widget = map_widget
        self._type = "legendWidget"

    @classmethod
    def _from_json(cls, widget_json):
        map_widget = widget_json["mapWidgetId"]
        name = widget_json["name"]
        title = widget_json["caption"]
        description = widget_json["description"]
        map_legend = MapLegend(map_widget, name, title, description)
        map_legend._id = widget_json["id"]
        map_legend.no_data.alignment = widget_json["noDataVerticalAlignment"]
        map_legend.no_data.show_title = widget_json["showCaptionWhenNoData"]
        map_legend.no_data.show_description = widget_json["showDescriptionWhenNoData"]

        return map_legend

    def _repr_html_(self):
        from arcgis.apps.dashboard import Dashboard
        url = Dashboard._publish_random(self)
        return f"""<iframe src={url} width=900 height=300>"""

    def _convert_to_json(self):
        data = {
            "type": "legendWidget",
            "mapWidgetId": self._map_widget._id,
            "id": self._id,
            "name": self.name,
            "caption": self._title,
            "description": self._description,
            "showLastUpdate": True,
            "noDataVerticalAlignment": "middle",
            "showCaptionWhenNoData": True,
            "showDescriptionWhenNoData": True
        }

        if self._background_color:
            data["backgroundColor"] = self._background_color

        if self._text_color:
            data["textColor"] = self._text_color

        return data



import traceback
from .._data import _raise_fastai_import_error
HAS_TRANSFORMER = True

try:
    import torch
    from transformers import pipeline, logging
    from .._utils.common import _get_device_id
    from fastprogress.fastprogress import progress_bar
    from transformers.modeling_auto import MODEL_FOR_SEQUENCE_CLASSIFICATION_MAPPING
    EXPECTED_MODEL_TYPES = [x.__name__.replace('Config', '') for x in MODEL_FOR_SEQUENCE_CLASSIFICATION_MAPPING.keys()]
except Exception as e:
    transformer_exception = "\n".join(traceback.format_exception(type(e), e, e.__traceback__))
    HAS_TRANSFORMER = False
    EXPECTED_MODEL_TYPES = []


class ZeroShotClassifier:
    """
    Creates a `ZeroShotClassifier` Object.
    Based on the Hugging Face transformers library

    =====================   ===========================================
    **Argument**            **Description**
    ---------------------   -------------------------------------------
    backbone                Optional string. Specifying the HuggingFace
                            transformer model name which will be used to
                            predict the answers from a given passage/context.

                            To learn more about the available models for
                            zero-shot-classification task, kindly visit:-
                            https://huggingface.co/models?search=nli
    =====================   ===========================================

    :returns: `ZeroShotClassifier` Object
    """

    #: supported transformer backbones
    supported_backbones = EXPECTED_MODEL_TYPES

    def __init__(self, backbone=None):
        if not HAS_TRANSFORMER:
            _raise_fastai_import_error(import_exception=transformer_exception)

        logger = logging.get_logger()
        logger.setLevel(logging.ERROR)
        self._device = _get_device_id()
        self._task = "zero-shot-classification"
        try:
            self.model = pipeline(self._task, model=backbone, device=self._device)
        except Exception as e:
            error_message = (f"Model - `{backbone}` cannot be used for {self._task} task.\n"
                             f"Model type should be one of {EXPECTED_MODEL_TYPES}.")
            raise Exception(error_message)
        
    def predict(self, text_or_list, candidate_labels, **kwargs):
        """
        Predicts the class label(s) for the input text

        =====================   ===========================================
        **Argument**            **Description**
        ---------------------   -------------------------------------------
        text_or_list            Required string or list. The sequence or a
                                list of sequences to classify.
        ---------------------   -------------------------------------------
        candidate_labels        Required string or list. The set of possible
                                class labels to classify each sequence into.
                                Can be a single label, a string of
                                comma-separated labels, or a list of labels.
        =====================   ===========================================

        **kwargs**

        =====================   ===========================================
        **Argument**            **Description**
        ---------------------   -------------------------------------------
        multi_class             Optional boolean. Whether or not multiple
                                candidate labels can be true.
                                Default value is set to False.
        ---------------------   -------------------------------------------
        hypothesis              Optional string. The template used to turn each
                                label into an NLI-style hypothesis. This template
                                must include a {} or similar syntax for the
                                candidate label to be inserted into the template.
                                Default value is set to `"This example is {}."`.
        =====================   ===========================================

        :returns: a list of :obj:`dict`: Each result comes as a dictionary with the following keys:

            - **sequence** (:obj:`str`) -- The sequence for which this is the output.
            - **labels** (:obj:`List[str]`) -- The labels sorted by order of likelihood.
            - **scores** (:obj:`List[float]`) -- The probabilities for each of the labels.
        """
        results = []
        multi_class = kwargs.get("multi_class", False)
        hypothesis = kwargs.get("hypothesis", "This example is {}.")
        if not isinstance(text_or_list, (list, tuple)): text_or_list = [text_or_list]
        for i in progress_bar(range(len(text_or_list))):
            result = self.model(text_or_list[i], candidate_labels,
                                multi_class=multi_class, hypothesis_template=hypothesis)
            results.append(result)
        from IPython.display import clear_output
        clear_output()
        return results



""" Defines public exports for the workforce module.
"""

from .assignment import Assignment
from .assignment_type import AssignmentType
from .attachment import Attachment
from .dispatcher import Dispatcher
from .integration import Integration
from .exceptions import ServerError, ValidationError
from .project import Project
from .track import Track
from .worker import Worker
from . import _store
from ._store.projects import create_project



""" Defines store functions for working with Integrations.
"""

from ... import workforce
from .utils import add_features, update_features, remove_features, validate


def get_integration(project, integration_id=None):
    """ Gets the identified Integration.  Exactly one integration_id should be provided.
        :param project:
        :param integration_id
        :returns: Integration
    """
    if integration_id:
        where = "{} = '{}'".format(project._integration_schema.integration_id, integration_id)
    else:
        where = "1=0"
    integrations = query_integrations(project, where)
    return integrations if integrations else None


def query_integrations(project, where):
    """ Executes a query against the integrations table.
        :param project: The project in which to query integrations.
        :param where: An ArcGIS where clause.
        :returns: list of Integrations
    """
    features = project.integrations_table.query(where, return_all_records=True).features
    return [workforce.Integration(project, feature) for feature in features]


def add_integration(project, feature=None, integration_id=None, prompt=None, url_template=None, assignment_type=None):
    """
    Adds a new integration to the project
    """
    project._update_cached_objects()
    integration = workforce.Integration(project,
                                      feature,
                                      integration_id,
                                      prompt,
                                      url_template,
                                      assignment_type)
    return add_integrations(project, [integration])[0]


def add_integrations(project, integrations):
    """ Adds Integrations to a project.

        :param project:
        :param integrations: list of Integrations
        :raises ValidationError: Indicates that one or more integrations failed validation.
        :raises ServerError: Indicates that the server rejected the integrations.
    """
    project._update_cached_objects()
    if integrations:
        use_global_ids = True
        for integration in integrations:
            validate(integration._validate)
            if integration.global_id is None:
                use_global_ids = False
                break

        features = [integration.feature for integration in integrations]
        add_features(project.integrations_table, features, use_global_ids)

    return integrations


def update_integration(project, integration, integration_id=None, prompt=None, url_template=None, assignment_type=None):
    """
        Updates an integration and submits changes to the server
    """
    project._update_cached_objects()
    if integration_id:
        integration.integration_id = integration_id
    if prompt:
        integration.prompt = prompt
    if url_template:
        integration.url_template = url_template
    if assignment_type:
        # GUIDs are cast to upper for internal storage purposes
        integration.assignment_type = assignment_type.upper()
    return update_integrations(project, [integration])[0]


def update_integrations(project, integrations):
    """ Updates Integrations to a project.
        :param project:
        :param integrations: list of Integrations
        :raises ValidationError: Indicates that one or more dispatchers failed validation.
        :raises ServerError: Indicates that the server rejected the dispatchers.
    """
    project._update_cached_objects()
    if integrations:
        for integration in integrations:
            validate(integration._validate)
        features = [integration.feature for integration in integrations]
        update_features(project.integrations_table, features)
    
    return integrations


def delete_integrations(project, integrations):
    """ Removes Integrations from the project.
        :param project:
        :param integrations: list of Integrations
        :raises ValidationError: Indicates that one or more integrations failed validation.
        :raises ServerError: Indicates that the server rejected the removal.
    """
    project._update_cached_objects()
    if integrations:
        features = [integration.feature for integration in integrations]
        remove_features(project.integrations_table, features)



# Note that these may change and need to be updated periodically

basemap_dict = {'streets': [{'id': 'streets-base-layer',
                            'url': 'https://services.arcgisonline.com/ArcGIS/rest/services/World_Street_Map/MapServer',
                            'layerType': 'ArcGISTiledMapServiceLayer',
                            'title': 'World Street Map',
                            'visibility': True,
                            'opacity': 1
                            }],
                'satellite': [{'id': 'satellite-base-layer',
                            'url': 'https://services.arcgisonline.com/ArcGIS/rest/services/World_Imagery/MapServer',
                            'layerType': 'ArcGISTiledMapServiceLayer',
                            'title': 'World Imagery',
                            'visibility': True,
                            'opacity': 1
                            }],
                'hybrid': [{'id': 'hybrid-base-layer',
                            'url': 'https://services.arcgisonline.com/ArcGIS/rest/services/World_Imagery/MapServer',
                            'layerType': 'ArcGISTiledMapServiceLayer',
                            'title': 'World Imagery',
                            'visibility': True,
                            'opacity': 1
                            },
                           {'id': 'hybrid-reference-layer',
                            'url': 'https://services.arcgisonline.com/ArcGIS/rest/services/Reference/World_Boundaries_and_Places/MapServer',
                            'layerType': 'ArcGISTiledMapServiceLayer',
                            'title': 'World Boundaries and Places',
                            'isReference': True,
                            'visibility': True,
                            'opacity': 1
                            }
                           ],
                'terrain': [{'id': 'terrain-base-layer',
                            'url': 'https://services.arcgisonline.com/ArcGIS/rest/services/World_Terrain_Base/MapServer',
                            'layerType': 'ArcGISTiledMapServiceLayer',
                            'title': 'World Terrain',
                            'visibility': True,
                            'opacity': 1
                            },
                           {'id': 'terrain-reference-layer',
                            'url': 'https://services.arcgisonline.com/ArcGIS/rest/services/Reference/World_Reference_Overlay/MapServer',
                            'layerType': 'ArcGISTiledMapServiceLayer',
                            'title': 'World Reference Overlay',
                            'isReference': True,
                            'visibility': True,
                            'opacity': 1
                            }
                           ],
                'topo': [{'id': 'topo-base-layer',
                            'url': 'https://services.arcgisonline.com/ArcGIS/rest/services/World_Topo_Map/MapServer',
                            'layerType': 'ArcGISTiledMapServiceLayer',
                            'title': 'World Topo Map',
                            'visibility': True,
                            'opacity': 1
                            }],
                'gray': [{'id': 'gray-base-layer',
                            'url': 'https://services.arcgisonline.com/ArcGIS/rest/services/Canvas/World_Light_Gray_Base/MapServer',
                            'layerType': 'ArcGISTiledMapServiceLayer',
                            'title': 'World Light Gray Base',
                            'visibility': True,
                            'opacity': 1
                            },
                           {'id': 'gray-reference-layer',
                            'url': 'https://services.arcgisonline.com/ArcGIS/rest/services/Canvas/World_Light_Gray_Reference/MapServer',
                            'layerType': 'ArcGISTiledMapServiceLayer',
                            'title': 'World Light Gray Reference',
                            'isReference': True,
                            'visibility': True,
                            'opacity': 1
                            }
                           ],
                'dark-gray': [{'id': 'dark-gray-base-layer',
                            'url': 'https://services.arcgisonline.com/ArcGIS/rest/services/Canvas/World_Dark_Gray_Base/MapServer',
                            'layerType': 'ArcGISTiledMapServiceLayer',
                            'title': 'World Dark Gray Base',
                            'visibility': True,
                            'opacity': 1
                            },
                           {'id': 'dark-gray-reference-layer',
                            'url': 'https://services.arcgisonline.com/ArcGIS/rest/services/Canvas/World_Dark_Gray_Reference/MapServer',
                            'layerType': 'ArcGISTiledMapServiceLayer',
                            'title': 'World Dark Gray Reference',
                            'isReference': True,
                            'visibility': True,
                            'opacity': 1
                            }
                           ],
                'oceans': [{'id': 'oceans-base-layer',
                            'url': 'https://services.arcgisonline.com/arcgis/rest/services/Ocean/World_Ocean_Base/MapServer',
                            'layerType': 'ArcGISTiledMapServiceLayer',
                            'title': 'World Oceans Base',
                            'visibility': True,
                            'opacity': 1
                            },
                           {'id': 'oceans-reference-layer',
                            'url': 'https://services.arcgisonline.com/ArcGIS/rest/services/Ocean/World_Ocean_Reference/MapServer',
                            'layerType': 'ArcGISTiledMapServiceLayer',
                            'title': 'World Oceans Reference',
                            'isReference': True,
                            'visibility': True,
                            'opacity': 1
                            }
                           ],
                'national-geographic': [{'id': 'national-geographic-base-layer',
                            'url': 'https://services.arcgisonline.com/ArcGIS/rest/services/NatGeo_World_Map/MapServer',
                            'layerType': 'ArcGISTiledMapServiceLayer',
                            'title': 'NatGeo World Map',
                            'visibility': True,
                            'opacity': 1
                            }],
                'osm': [{'id': 'osm-base-layer',
                            'styleUrl': 'https://cdn.arcgis.com/sharing/rest/content/items/3e1a00aeae81496587988075fe529f71/resources/styles/root.json',
                            'layerType': 'OpenStreetMap',
                            'title': 'Open Street Map',
                            'visibility': True,
                            'opacity': 1
                            }],
                'dark-gray-vector': [{'id': 'dark-gray-base-layer',
                            'styleUrl': 'https://www.arcgis.com/sharing/rest/content/items/c11ce4f7801740b2905eb03ddc963ac8/resources/styles/root.json',
                            'layerType': 'VectorTileLayer',
                            'title': 'World Dark Gray',
                            'visibility': True,
                            'opacity': 1
                            }],
                'gray-vector': [{'id': 'gray-base-layer',
                            'styleUrl': 'https://www.arcgis.com/sharing/rest/content/items/8a2cba3b0ebf4140b7c0dc5ee149549a/resources/styles/root.json',
                            'layerType': 'VectorTileLayer',
                            'title': 'World Light Gray',
                            'visibility': True,
                            'opacity': 1
                            }],
                'streets-vector': [{'id': 'streets-vector-base-layer',
                            'styleUrl': 'https://www.arcgis.com/sharing/rest/content/items/de26a3cf4cc9451298ea173c4b324736/resources/styles/root.json',
                            'layerType': 'VectorTileLayer',
                            'title': 'World Light Gray',
                            'visibility': True,
                            'opacity': 1
                            }],
                'topo-vector': [{'id': 'world-hillshade-layer',
                            'url': 'https://services.arcgisonline.com/arcgis/rest/services/Elevation/World_Hillshade/MapServer',
                            'layerType': 'ArcGISTiledMapServiceLayer',
                            'title': 'World Hillshade',
                            'visibility': True,
                            'opacity': 1},
                            {'id': 'topo-vector-base-layer',
                              'styleUrl': 'https://www.arcgis.com/sharing/rest/content/items/7dc6cea0b1764a1f9af2e679f642f0f5/resources/styles/root.json',
                              'layerType': 'VectorTileLayer',
                              'title': 'World Topo',
                              'visibility': True,
                              'opacity': 1
                            }],
                'streets-night-vector': [{'id': 'streets-night-vector-base-layer',
                            'styleUrl': 'https://www.arcgis.com/sharing/rest/content/items/86f556a2d1fd468181855a35e344567f/resources/styles/root.json',
                            'layerType': 'VectorTileLayer',
                            'title': 'World Streets Night',
                            'visibility': True,
                            'opacity': 1
                            }],
                'streets-relief-vector': [{'id': 'world-hillshade-layer',
                            'url': 'https://services.arcgisonline.com/arcgis/rest/services/Elevation/World_Hillshade/MapServer',
                            'layerType': 'ArcGISTiledMapServiceLayer',
                            'title': 'World Hillshade',
                            'visibility': True,
                            'opacity': 1},
                            {'id': 'streets-relief-vector-base-layer',
                              'styleUrl': 'https://www.arcgis.com/sharing/rest/content/items/b266e6d17fc345b498345613930fbd76/resources/styles/root.json',
                              'layerType': 'VectorTileLayer',
                              'title': 'World Streets Relief',
                              'visibility': True,
                              'opacity': 1
                            }],
                'streets-navigation-vector': [{'id': 'streets-navigation-vector-base-layer',
                            'styleUrl': 'https://www.arcgis.com/sharing/rest/content/items/63c47b7177f946b49902c24129b87252/resources/styles/root.json',
                            'layerType': 'VectorTileLayer',
                            'title': 'World Streets Navigation',
                            'visibility': True,
                            'opacity': 1
                            }]
                }


import os
import traceback
import json
import math
import warnings
from zipfile import ZipFile
import tempfile
from pathlib import Path
HAS_FASTAI = False
try:
    from .env import raise_fastai_import_error, HAS_GDAL, gdal_import_exception, GDAL_INSTALL_MESSAGE, LAMBDA_TEXT_CLASSIFICATION
    if not LAMBDA_TEXT_CLASSIFICATION:
        from fastai.vision.data import ImageList
        from fastai.vision import Image, imagenet_stats, pil2tensor
        import PIL
    elif LAMBDA_TEXT_CLASSIFICATION:
        missing_classes = ['Image', 'ImageList']
        for missing_class in missing_classes: 
            exec(f'{missing_class} = object')
    from fastai.data_block import get_files
    import torch
    import numpy as np
    from matplotlib import pyplot as plt
    HAS_FASTAI = True
except Exception:
    import_exception = traceback.format_exc()
    pass


def read_image(path, resize_to: int=None):
    """
    path: file path of image on disk.

    returns an array of image pixel values in channel last format.
    """
    path = str(os.path.abspath(path))
    if not os.path.exists:
        raise Exception(f"The image path {path} could not be found on disk, please verify your training data.")

    gdal_error = None
    try:
        if not HAS_GDAL:
            gdal_error = f"""{gdal_import_exception} \n\n{GDAL_INSTALL_MESSAGE}"""
        else:
            from osgeo import gdal
            ds = gdal.Open(path)
            if resize_to is None:
                arr = ds.ReadAsArray()
            else:
                gdal_dtype = ds.GetRasterBand(1).DataType
                transform = ds.GetGeoTransform()
                dx = transform[1]
                dy = transform[-1]
                #
                dx_new = (ds.RasterXSize / resize_to) * dx
                dy_new = (ds.RasterYSize / resize_to) * dy
                #
                ds_new = gdal.Warp(
                    '',
                    path,
                    dstSRS=ds.GetProjection(),
                    format='VRT',
                    outputType=gdal_dtype,
                    xRes=dx_new,
                    yRes=dy_new
                )
                arr = ds_new.ReadAsArray()
            if len(arr.shape) > 2:
                arr = np.rollaxis(arr, 0, 3)
            return arr
    except Exception as _gdal_error:
        gdal_error = str(_gdal_error)

    # Attach gdal error
    message = f"""
       Tried opening image using gdal and encountered the following error
       \n\n{gdal_error}
       """

    if resize_to is not None:
        raise Exception(f"`resize_to` parameter is only supported using gdal. \n"+message)

    try:
        from skimage.io import imread
        return np.array(PIL.Image.open(path).convert('RGB'))
    except Exception as _pillow_error:
        pillow_error = str(_pillow_error)

    # Attach pillow error
    message += f"""
       \n===================================================================
       \n\nTried opening image using pillow and encountered the following error
       \n\n{pillow_error}
       """
    raise Exception(message)

class ArcGISMSImage(Image):

    def show(self, ax=None, rgb_bands=None, show_axis=False, title=None):
        if rgb_bands is None:
            rgb_bands = getattr(self, 'rgb_bands', [0, 1, 2])
        if ax is None:
            ax = plt.subplot(1, 1, 1)
        symbology_data = self.data[rgb_bands]
        im_shape = symbology_data.shape
        min_vals = symbology_data.view(im_shape[0], -1).min(dim=1)[0]
        max_vals = symbology_data.view(im_shape[0], -1).max(dim=1)[0]
        strechted_data = ( symbology_data - min_vals.view(im_shape[0], 1, 1) ) / ( max_vals.view(im_shape[0], 1, 1) - min_vals.view(im_shape[0], 1, 1) + .001 )
        data_to_plot = strechted_data.permute(1, 2, 0)
        if not show_axis:
            ax.axis('off')
        ax.imshow(data_to_plot)
        if title is not None:
            ax.set_title(title)

    def print_method(self):
        return self.show()

    def _repr_png_(self):
        return self.show()

    def _repr_jpeg_(self):
        return self.show()

    @classmethod
    def open_gdal(cls, path):
        try:
            from osgeo import gdal
        except ImportError as e:
            message = f"""
            {e}\n\nPlease install gdal using the following command
            \nconda install gdal=2.3.3
            """
            raise Exception(message)
        path = str(os.path.abspath(path))
        x = gdal.Open(path).ReadAsArray()
        x = torch.tensor(x.astype(np.float32))
        if len(x.shape)==2:
            x = x.unsqueeze(0)
        return cls(x)

    @staticmethod
    def read_image(path):
        return read_image(path)

    @classmethod
    def open(cls, path, cast_to=np.float32, div=None, imagery_type=None):
        path = str(os.path.abspath(path))
        if not os.path.exists:
            raise Exception(f"The image path {path} could not be found on disk, please verify your training data.")

        read = False
        gdal_error = None
        pillow_error = None
        try:
            from osgeo import gdal
            x = gdal.Open(path).ReadAsArray()
            # Ignore Alpha Channel
            if x.shape[0] == 4 and imagery_type == 'RGB':
                x = x[:3]
            x = torch.tensor(x.astype(cast_to))
            read = True
        except Exception as _gdal_error:
            gdal_error = str(_gdal_error)
            pass

        if not read:
            try:
                with warnings.catch_warnings():
                    warnings.simplefilter("ignore", UserWarning)  # EXIF warning from TiffPlugin
                    x = PIL.Image.open(path).convert('RGB')
                x = pil2tensor(x, cast_to)
                read = True
            except Exception as _pillow_error:
                pillow_error = str(_pillow_error)
                pass

        if not read:
            # Attach gdal error
            message = f"""
            Tried opening image using gdal and encountered the following error
            \n\n{gdal_error}
            """
            if (gdal_error) == ModuleNotFoundError:
                message += GDAL_INSTALL_MESSAGE
            # Attach pillow error
            message += f"""
            \n===================================================================
            \n\nTried opening image using pillow and encountered the following error
            \n\n{pillow_error}
            """

            raise Exception(message)
        else:
            if len(x.shape)==2:
                x = x.unsqueeze(0)
            if div is not None:
                x = x / div

        return cls(x)

class ArcGISImageList(ImageList):
    "`ImageList` suitable for classification tasks."
    _square_show_res = False
    _div = None
    _imagery_type = None
    def open(self, fn):
        return ArcGISMSImage.open(fn, div=self._div, imagery_type=self._imagery_type)

def get_multispectral_data_params_from_emd(data, emd):
    data._is_multispectral = emd.get('IsMultispectral', False)
    if data._is_multispectral:
        data._bands = emd.get('Bands')
        data._imagery_type = emd.get("ImageryType")
        data._extract_bands = emd.get("ExtractBands")
        data._train_tail = False # Hardcoded because we are never going to train a model with empty data
        normalization_stats = dict(emd.get("NormalizationStats")) # Copy the normalization stats so that self._data.emd has no tensors other wise it will raise error while creating emd
        for _stat in normalization_stats:
            if normalization_stats[_stat] is not None:
                normalization_stats[_stat] = torch.tensor(normalization_stats[_stat])
            setattr(data, ('_'+_stat), normalization_stats[_stat])
        data._do_normalize = emd.get("DoNormalize")
    return data

def get_post_processed_model(arcgis_model, input_normalization=True):
    if arcgis_model._backend == 'tensorflow':
        from .postprocessing_tf import get_post_processed_model_tf
        return get_post_processed_model_tf(arcgis_model, input_normalization=input_normalization)

def get_color_array(color_mapping: dict, alpha=0.7):
    color_array = np.array(list(color_mapping.values()), dtype=np.float) / 255
    color_array = np.concatenate([color_array, np.repeat([alpha], color_array.shape[0]).reshape(color_array.shape[0], 1)], axis=-1)
    return color_array

## show_batch() show_results() helper functions start ##

def to_torch_tensor(x):
    if type(x) == torch.Tensor:
        return x.clone().detach()
    return torch.tensor(x)

def get_symbology_bands(rgb_bands, extract_bands, bands):
    e = Exception('`rgb_bands` should be a valid band_order, list or tuple of length 3 or 1.')
    symbology_bands = []
    if not ( len(rgb_bands) == 3 or len(rgb_bands) == 1 ):
        raise(e)
    for b in rgb_bands:
        if type(b) == str:
            b_index = bands.index(b)
        elif type(b) == int:
            # check if the band index specified by the user really exists.
            bands[b]
            b_index = b
        else:
            raise(e)
        b_index = extract_bands.index(b_index)
        symbology_bands.append(b_index)
    return symbology_bands

def get_top_padding(title_font_size, nrows, imsize):
    top = 1 - (math.sqrt(title_font_size)/math.sqrt(100*nrows*imsize))
    return top

def kwarg_fill_none(kwargs, kwarg_key, default_value=None):
    value = kwargs.get(kwarg_key, None)
    if value is None:
        return default_value
    return value

def find_data_loader(type_data_loader, data):
    if type_data_loader == 'training':
        data_loader = data.train_dl
    elif type_data_loader == 'validation':
        data_loader = data.valid_dl
    elif type_data_loader == 'testing':
        data_loader = data.test_dl
    else:
        e = Exception(f'could not find {type_data_loader} in data.')
        raise(e)
    return data_loader

def get_nbatches(data_loader, nbatches):
    x_batch, y_batch = [], []
    dl_iterater = iter(data_loader)
    get_next = True
    i = 0
    while i < nbatches and get_next:
        i+=1
        try:
            x, y = next(dl_iterater)
            x_batch.append(x)
            y_batch.append(y)
        except StopIteration:
            get_next = False
    return x_batch, y_batch

def image_tensor_checks_plotting(imagetensor_batch):
    symbology_x_batch = imagetensor_batch
    # Channel first to channel last for plotting
    symbology_x_batch = symbology_x_batch.permute(0, 2, 3, 1)

    # Clamp float values to range 0 - 1
    if symbology_x_batch.mean() < 1:
        symbology_x_batch = symbology_x_batch.clamp(0, 1)

    # Squeeze channels if single channel (1, 224, 224) -> (224, 224)
    if symbology_x_batch.shape[-1] == 1:
        symbology_x_batch = symbology_x_batch.squeeze(-1)
    return symbology_x_batch

def denorm_x(imagetensor_batch, self=None):
    """
    denormalizes a imagetensor_batch for plotting
    -------------------------
    imagetensor_batch: imagebatch with shape (batch, bands, rows, columns)

    self: optional. can be an instance of ArcGISModel or arcgis.learn data object(databunch)
    -------------------------
    returns denormalized imagetensor_batch
    """
    from .. import models
    if isinstance(self, models._arcgis_model.ArcGISModel):
        data = self._data
    else:
        data = self
    if data is not None and data._is_multispectral:
        return denorm_image(
            imagetensor_batch,
            mean=data._scaled_mean_values[data._extract_bands],
            std=data._scaled_std_values[data._extract_bands]
        )
    return denorm_image(imagetensor_batch)

def denorm_image(imagetensor_batch, mean=None, std=None):
    # prepare normalization stats
    if mean is None or std is None:
        mean = imagenet_stats[0]
        std = imagenet_stats[1]
    mean = to_torch_tensor(mean).to(imagetensor_batch).view(1, -1, 1, 1)
    std = to_torch_tensor(std).to(imagetensor_batch).view(1, -1, 1, 1)
    return (imagetensor_batch * std) + mean

def predict_batch(self, imagetensor_batch):
    if self._backend == 'pytorch':
        predictions = self.learn.model.eval()(imagetensor_batch.to(self._device).float()).detach()
        return predictions
    elif self._backend == 'tensorflow':
        from .common_tf import predict_batch_tf
        return predict_batch_tf(self, imagetensor_batch)

## show_batch() show_results() helper functions end ##

## Image Stretching Functions start ##

def get_band_percent_minmax(values, min_clip, max_clip):
    return values[round(values.shape[0] * min_clip)], values[values.shape[0] - round(values.shape[0] * max_clip)]

def get_percent_minmax(imagetensor_batch, min_clip=0.0025, max_clip=0.005):
    shp = imagetensor_batch.shape
    _imagetensor_batch = imagetensor_batch.transpose(1, 0).reshape(shp[1], -1)
    min_vals = []
    max_vals = []
    for i in range(shp[1]):
        v = get_band_percent_minmax(_imagetensor_batch[i].unique(), min_clip, max_clip)
        min_vals.append(v[0])
        max_vals.append(v[1])
    return \
        torch.tensor(
            min_vals,
            dtype=imagetensor_batch.dtype,
            device=imagetensor_batch.device
        ),\
        torch.tensor(
            max_vals,
            dtype=imagetensor_batch.dtype,
            device=imagetensor_batch.device
        )

def image_batch_stretcher(imagetensor_batch, stretch_type='minmax', statistics_type=None):
    shp = imagetensor_batch.shape
    if statistics_type == 'DRA':
        if stretch_type == 'minmax':
            min_vals = imagetensor_batch.view(shp[0], shp[1], -1).min(dim=2)[0]
            max_vals = imagetensor_batch.view(shp[0], shp[1], -1).max(dim=2)[0]
        elif stretch_type == 'percentclip':
            min_vals = []
            max_vals = []
            for idx in range(shp[0]):
                v = get_percent_minmax(imagetensor_batch[idx:idx+1])
                min_vals.append(v[0])
                max_vals.append(v[1])
            min_vals = torch.stack(min_vals)
            max_vals = torch.stack(max_vals)
        else:
            raise NotImplementedError
        min_vals = min_vals.view(shp[0], shp[1], 1, 1)
        max_vals = max_vals.view(shp[0], shp[1], 1, 1)
    else:
        if stretch_type == 'minmax':
            min_vals = imagetensor_batch.transpose(1, 0).reshape(shp[1], -1).min(dim=1)[0]
            max_vals = imagetensor_batch.transpose(1, 0).reshape(shp[1], -1).max(dim=1)[0]
        elif stretch_type == 'percentclip':
            min_vals, max_vals = get_percent_minmax(imagetensor_batch)
        else:
            raise NotImplementedError
        min_vals = min_vals.view(1, shp[1], 1, 1)
        max_vals = max_vals.view(1, shp[1], 1, 1)
    #
    imagetensor_batch = ( imagetensor_batch - min_vals ) / ( max_vals - min_vals + .001 )
    imagetensor_batch = imagetensor_batch.clamp(0, 1)
    return imagetensor_batch

def dynamic_range_adjustment(imagetensor_batch):
    shp = imagetensor_batch.shape
    min_vals = imagetensor_batch.view(shp[0], shp[1], -1).min(dim=2)[0]
    max_vals = imagetensor_batch.view(shp[0], shp[1], -1).max(dim=2)[0]
    imagetensor_batch = ( imagetensor_batch - min_vals.view(shp[0], shp[1], 1, 1) ) / ( max_vals.view(shp[0], shp[1], 1, 1) - min_vals.view(shp[0], shp[1], 1, 1) + .001 )
    imagetensor_batch = imagetensor_batch.clamp(0, 1)
    return imagetensor_batch

## Image Stretching Functions end ##

def load_model(emd_path, data=None):
    from .. import models
    # if not HAS_FASTAI:
    #     raise_fastai_import_error(import_exception=import_exception)

    _emd_path = os.path.abspath(emd_path)
    if not os.path.exists(emd_path):
        raise Exception(f"Could not find an EMD file at the specified path does not exist '{emd_path}'")

    with open(_emd_path) as f:
        emd = json.load(f)
    model_name = emd['ModelName']
    model_cls = getattr(models, model_name, None)

    if model_cls is None:
        raise Exception(f"Failed to load model, Could not find class '{model_name}' in arcgis.learn.models")

    model_obj = model_cls.from_model(_emd_path, data=data)

    return model_obj


def _temp_dlpk(dlpk_path):
    with ZipFile(dlpk_path, 'r') as zip_obj:
        temp_dir = tempfile.TemporaryDirectory().name
        zip_obj.extractall(temp_dir)
    return temp_dir


def _get_emd_path(emd_path):
    emd_path = Path(emd_path)
    if emd_path.suffix == '.dlpk':
        temp_path = _temp_dlpk(emd_path)
        emd_path = Path(temp_path)
        #return cls.from_model(temp_path)

    if emd_path.suffix != '.emd':
        list_files = get_files(emd_path, extensions=['.emd'])
        assert(len(list_files)==1)
        #return cls.from_model(list_files[0])
        emd_path = list_files[0]
    return emd_path


def _get_gpu_device_id(max_memory=0.8):
    '''
    select available device based on the memory utilization status of the device
    :param max_memory: the maximum memory utilization ratio that is considered available
    :return: GPU id that is available, -1 means no GPU is available/uses CPU, if GPUtil package is not installed, will
    return 0
    '''
    try:
        import GPUtil
    except ModuleNotFoundError:
        return 0

    GPUs = GPUtil.getGPUs()
    freeMemory, available = 0, 0
    for GPU in GPUs:
        if GPU.memoryUtil > max_memory:
            continue
        if GPU.memoryFree >= freeMemory:
            freeMemory = GPU.memoryFree
            available = GPU.id

    return available


def _get_device_id():
    import arcgis
    from ..models._arcgis_model import _device_check
    move_to_cpu = _device_check()
    if move_to_cpu: arcgis.env._processorType = "CPU"

    if getattr(arcgis.env, "_processorType", "") == "GPU" and torch.cuda.is_available():
        device = _get_gpu_device_id()
    elif getattr(arcgis.env, "_processorType", "") == "CPU":
        device = -1
    else:
        device = _get_gpu_device_id() if torch.cuda.is_available() else -1

    return device



import traceback
from .._data import _raise_fastai_import_error
HAS_TRANSFORMER = True

try:
    import torch
    from transformers import pipeline, logging
    from .._utils.common import _get_device_id
    from fastprogress.fastprogress import progress_bar
    from transformers.modeling_auto import MODEL_FOR_MASKED_LM_MAPPING
    EXPECTED_MODEL_TYPES = [x.__name__.replace('Config', '') for x in MODEL_FOR_MASKED_LM_MAPPING.keys()]
except Exception as e:
    transformer_exception = "\n".join(traceback.format_exception(type(e), e, e.__traceback__))
    HAS_TRANSFORMER = False
    EXPECTED_MODEL_TYPES = []

class FillMask:
    """
    Creates a `FillMask` Object.
    Based on the Hugging Face transformers library

    =====================   ===========================================
    **Argument**            **Description**
    ---------------------   -------------------------------------------
    backbone                Optional string. Specify the HuggingFace
                            transformer model name which will be used to
                            generate the suggestion token.

                            To learn more about the available models for
                            fill-mask task, kindly visit:-
                            https://huggingface.co/models?filter=lm-head
    =====================   ===========================================

    :returns: `FillMask` Object
    """

    #: supported transformer backbones
    supported_backbones = EXPECTED_MODEL_TYPES

    def __init__(self, backbone=None):
        if not HAS_TRANSFORMER:
            _raise_fastai_import_error(import_exception=transformer_exception)

        logger = logging.get_logger()
        logger.setLevel(logging.ERROR)
        self._device = _get_device_id()
        self._task = "fill-mask"
        try:
            self.model = pipeline(self._task, model=backbone, device=self._device, topk=10)
        except Exception as e:
            error_message = (f"Model - `{backbone}` cannot be used for {self._task} task.\n"
                             f"Model type should be one of {EXPECTED_MODEL_TYPES}.")
            raise Exception(error_message)

    def predict_token(self, text_or_list, num_suggestions=5):
        """
        Summarize the given text or list of text

        =====================   ===========================================
        **Argument**            **Description**
        ---------------------   -------------------------------------------
        text_or_list            Required string or list. A text/sentence
                                or a list of texts/sentences for which on wishes
                                to generate the recommendations for masked-token.
        =====================   ===========================================

        :returns: A list or a list of list of :obj:`dict`: Each result comes as list of dictionaries with the following keys:

            - **sequence** (:obj:`str`)  -- The corresponding input with the mask token prediction.
            - **score** (:obj:`float`)   -- The corresponding probability.
            - **token_str** (:obj:`str`) -- The predicted token (to replace the masked one).
        """
        results = []
        if not isinstance(text_or_list, (list, tuple)): text_or_list = [text_or_list]
        self._do_sanity(text_or_list)
        for i in progress_bar(range(len(text_or_list))):
            text = text_or_list[i].replace('__', self.model.tokenizer.mask_token)
            result = self._process_result(self.model(text)[:num_suggestions])
            if num_suggestions == 1: result = result[0]
            results.append(result)
        return results

    @staticmethod
    def _do_sanity(text_list):
        for text in text_list:
            if '__' not in text:
                error_message = (
                    f"Text - `{text}` is mising `__` token. Use the `__` token to "
                    "specify where you want to generate the suggestions in the text."
                )
                raise Exception(error_message)

    def _process_result(self, result_list):
        for item in result_list:
            _ = item.pop("token", -1)
            token = item["token_str"]
            sequence = item["sequence"]
            item["token_str"] = self.model.tokenizer.convert_tokens_to_string(token).strip()
            item["sequence"] = self.model.tokenizer.decode(
                self.model.tokenizer.encode(sequence, add_special_tokens=False), skip_special_tokens=True)

        return result_list
